'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import (
    add_arc_from_to,
    remove_transition,
)
from pm4py.util import nx_utils


def remove_rendundant_invisible_transitions(net):
    """
    Remove redundant transitions from Petri net

    Parameters
    -----------
    net
        Petri net

    Returns
    -----------
    net
        Cleaned net
    """
    trans = [x for x in list(net.transitions) if not x.label]
    i = 0
    while i < len(trans):
        if trans[i] in net.transitions:
            preset_i = set(x.source for x in trans[i].in_arcs)
            postset_i = set(x.target for x in trans[i].out_arcs)
            j = 0
            while j < len(trans):
                if not j == i:
                    preset_j = set(x.source for x in trans[j].in_arcs)
                    postset_j = set(x.target for x in trans[j].out_arcs)
                    if len(preset_j) == len(preset_i) and len(postset_j) < len(
                        postset_i
                    ):
                        if len(preset_j.intersection(preset_i)) == len(
                            preset_j
                        ) and len(postset_j.intersection(postset_i)) == len(
                            postset_j
                        ):
                            remove_transition(net, trans[j])
                            del trans[j]
                            continue
                j = j + 1
            i = i + 1
    return net


def find_bindings(and_measures):
    """
    Find the bindings given the AND measures

    Parameters
    -------------
    and_measures
        AND measures

    Returns
    -------------
    bindings
        Bindings
    """
    G = nx_utils.Graph()
    allocated_nodes = set()
    for n1 in list(and_measures.keys()):
        if n1 not in allocated_nodes:
            allocated_nodes.add(n1)
            G.add_node(n1)
        for n2 in list(and_measures[n1].keys()):
            if n2 not in allocated_nodes:
                allocated_nodes.add(n2)
                G.add_node(n1)
            G.add_edge(n1, n2)
    ret = list(nx_utils.find_cliques(G))
    return ret


def apply(heu_net, parameters=None):
    """
    Converts an Heuristics Net to a Petri net

    Parameters
    --------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    if parameters is None:
        parameters = {}
    net = PetriNet("")
    im = Marking()
    fm = Marking()
    source_places = []
    sink_places = []
    hid_trans_count = 0
    for index, sa_list in enumerate(heu_net.start_activities):
        source = PetriNet.Place("source" + str(index))
        source_places.append(source)
        net.places.add(source)
        im[source] = 1
    for index, ea_list in enumerate(heu_net.end_activities):
        sink = PetriNet.Place("sink" + str(index))
        sink_places.append(sink)
        net.places.add(sink)
        fm[sink] = 1
    act_trans = {}
    who_is_entering = {}
    who_is_exiting = {}
    for act1_name in heu_net.nodes:
        act1 = heu_net.nodes[act1_name]
        if act1_name not in act_trans:
            act_trans[act1_name] = PetriNet.Transition(act1_name, act1_name)
            net.transitions.add(act_trans[act1_name])
            who_is_entering[act1_name] = set()
            who_is_exiting[act1_name] = set()
            for index, sa_list in enumerate(heu_net.start_activities):
                if act1_name in sa_list:
                    who_is_entering[act1_name].add((None, index))
            for index, ea_list in enumerate(heu_net.end_activities):
                if act1_name in ea_list:
                    who_is_exiting[act1_name].add((None, index))
        for act2 in act1.output_connections:
            act2_name = act2.node_name
            if act2_name not in act_trans:
                act_trans[act2_name] = PetriNet.Transition(
                    act2_name, act2_name
                )
                net.transitions.add(act_trans[act2_name])
                who_is_entering[act2_name] = set()
                who_is_exiting[act2_name] = set()
                for index, sa_list in enumerate(heu_net.start_activities):
                    if act2_name in sa_list:
                        who_is_entering[act2_name].add((None, index))
                for index, ea_list in enumerate(heu_net.end_activities):
                    if act2_name in ea_list:
                        who_is_exiting[act2_name].add((None, index))
            who_is_entering[act2_name].add((act1_name, None))
            who_is_exiting[act1_name].add((act2_name, None))
    places_entering = {}
    for act1 in who_is_entering:
        cliques = find_bindings(heu_net.nodes[act1].and_measures_in)
        places_entering[act1] = {}
        entering_activities = list(who_is_entering[act1])
        entering_activities_wo_source = sorted(
            [x for x in entering_activities if x[0] is not None],
            key=lambda x: x[0],
        )
        entering_activities_only_source = [
            x for x in entering_activities if x[0] is None
        ]
        if entering_activities_wo_source:
            master_place = PetriNet.Place("pre_" + act1)
            net.places.add(master_place)
            add_arc_from_to(master_place, act_trans[act1], net)
            if len(entering_activities) == 1:
                places_entering[act1][entering_activities[0]] = master_place
            else:
                for index, act in enumerate(entering_activities_wo_source):
                    if act[0] in heu_net.nodes[act1].and_measures_in:
                        z = 0
                        while z < len(cliques):
                            if act[0] in cliques[z]:
                                hid_trans_count = hid_trans_count + 1
                                hid_trans = PetriNet.Transition(
                                    "hid_" + str(hid_trans_count), None
                                )
                                net.transitions.add(hid_trans)
                                add_arc_from_to(hid_trans, master_place, net)
                                for act2 in cliques[z]:
                                    if (act2, None) not in places_entering[
                                        act1
                                    ]:
                                        s_place = PetriNet.Place(
                                            "splace_in_"
                                            + act1
                                            + "_"
                                            + act2
                                            + "_"
                                            + str(index)
                                        )
                                        net.places.add(s_place)
                                        places_entering[act1][
                                            (act2, None)
                                        ] = s_place
                                    add_arc_from_to(
                                        places_entering[act1][(act2, None)],
                                        hid_trans,
                                        net,
                                    )
                                del cliques[z]
                                continue
                            z = z + 1
                        pass
                    elif act not in places_entering[act1]:
                        hid_trans_count = hid_trans_count + 1
                        hid_trans = PetriNet.Transition(
                            "hid_" + str(hid_trans_count), None
                        )
                        net.transitions.add(hid_trans)
                        add_arc_from_to(hid_trans, master_place, net)
                        if act not in places_entering[act1]:
                            s_place = PetriNet.Place(
                                "splace_in_" + act1 + "_" + str(index)
                            )
                            net.places.add(s_place)
                            places_entering[act1][act] = s_place
                        add_arc_from_to(
                            places_entering[act1][act], hid_trans, net
                        )
        for el in entering_activities_only_source:
            if len(entering_activities) == 1:
                add_arc_from_to(source_places[el[1]], act_trans[act1], net)
            else:
                hid_trans_count = hid_trans_count + 1
                hid_trans = PetriNet.Transition(
                    "hid_" + str(hid_trans_count), None
                )
                net.transitions.add(hid_trans)
                add_arc_from_to(source_places[el[1]], hid_trans, net)
                add_arc_from_to(hid_trans, master_place, net)
    for act1 in who_is_exiting:
        cliques = find_bindings(heu_net.nodes[act1].and_measures_out)
        exiting_activities = list(who_is_exiting[act1])
        exiting_activities_wo_sink = sorted(
            [x for x in exiting_activities if x[0] is not None],
            key=lambda x: x[0],
        )
        exiting_activities_only_sink = [
            x for x in exiting_activities if x[0] is None
        ]
        if exiting_activities_wo_sink:
            if (
                len(exiting_activities) == 1
                and len(exiting_activities_wo_sink) == 1
            ):
                ex_act = exiting_activities_wo_sink[0]
                if (act1, None) in places_entering[ex_act[0]]:
                    add_arc_from_to(
                        act_trans[act1],
                        places_entering[ex_act[0]][(act1, None)],
                        net,
                    )
            else:
                int_place = PetriNet.Place("intplace_" + str(act1))
                net.places.add(int_place)
                add_arc_from_to(act_trans[act1], int_place, net)
                for ex_act in exiting_activities_wo_sink:
                    if (act1, None) in places_entering[ex_act[0]]:
                        if ex_act[0] in heu_net.nodes[act1].and_measures_out:
                            z = 0
                            while z < len(cliques):
                                if ex_act[0] in cliques[z]:
                                    hid_trans_count = hid_trans_count + 1
                                    hid_trans = PetriNet.Transition(
                                        "hid_" + str(hid_trans_count), None
                                    )
                                    net.transitions.add(hid_trans)
                                    add_arc_from_to(int_place, hid_trans, net)
                                    for act in cliques[z]:
                                        add_arc_from_to(
                                            hid_trans,
                                            places_entering[act][(act1, None)],
                                            net,
                                        )
                                    del cliques[z]
                                    continue
                                z = z + 1
                        else:
                            hid_trans_count = hid_trans_count + 1
                            hid_trans = PetriNet.Transition(
                                "hid_" + str(hid_trans_count), None
                            )
                            net.transitions.add(hid_trans)
                            add_arc_from_to(int_place, hid_trans, net)
                            add_arc_from_to(
                                hid_trans,
                                places_entering[ex_act[0]][(act1, None)],
                                net,
                            )
        for el in exiting_activities_only_sink:
            if len(exiting_activities) == 1:
                add_arc_from_to(act_trans[act1], sink_places[el[1]], net)
            else:
                hid_trans_count = hid_trans_count + 1
                hid_trans = PetriNet.Transition(
                    "hid_" + str(hid_trans_count), None
                )
                net.transitions.add(hid_trans)
                add_arc_from_to(int_place, hid_trans, net)
                add_arc_from_to(hid_trans, sink_places[el[1]], net)
    net = remove_rendundant_invisible_transitions(net)
    from pm4py.objects.petri_net.utils import reduction

    reduction.apply_simple_reduction(net)
    return net, im, fm



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, List, Union

import pandas as pd

from pm4py.statistics.overlap.utils import compute
from pm4py.util import constants, xes_constants, exec_utils


class Parameters(Enum):
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY


def apply(
    df: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[int]:
    """
    Counts the intersections of each interval event with the other interval events of the log
    (all the events are considered, not looking at the activity)

    Parameters
    ----------------
    df
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.START_TIMESTAMP_KEY => the attribute to consider as start timestamp
        - Parameters.TIMESTAMP_KEY => the attribute to consider as timestamp

    Returns
    -----------------
    overlap
        For each interval event, ordered by the order of appearance in the log, associates the number
        of intersecting events.
    """
    if parameters is None:
        parameters = {}

    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    df = df[list({start_timestamp_key, timestamp_key})].to_dict("records")
    points = []

    for event in df:
        points.append(
            (
                event[start_timestamp_key].timestamp(),
                event[timestamp_key].timestamp(),
            )
        )

    return compute.apply(points)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid
import numpy as np
import math
import copy
from typing import Optional, Dict, Any, List, Tuple, Set
from graphviz import Digraph
from enum import Enum
from pm4py.util import exec_utils, constants, vis_utils
from pm4py.util.lp import solver as lp_solver
from pm4py.objects.petri_net.obj import PetriNet  # For type hinting if needed
from copy import deepcopy


# --- Constants matching TypeScript OCPNLayout ---
PLACE_TYPE = "place"
TRANSITION_TYPE = "transition"
DUMMY_TYPE = "dummy"


# --- Helper Functions ---

def ot_to_color(ot: str) -> str:
    """Generates a deterministic hex color based on the object type string."""
    import hashlib
    hash_obj = hashlib.md5(ot.encode('utf-8'))
    hash_hex = hash_obj.hexdigest()
    color = "#" + hash_hex[:6]
    return color


def generate_dummy_id() -> str:
    """Generates a unique ID for a dummy node."""
    return "dummy_" + str(uuid.uuid4())


def generate_arc_id(source_id: str, target_id: str) -> str:
    """Generates a unique ID for an arc."""
    return f"arc_{source_id}_{target_id}_{uuid.uuid4()}"


def get_neighbors(vertex_id: str, layout_data: Dict, direction: str) -> List[str]:
    """Gets upper ('up') or lower ('down') neighbors of a vertex."""
    neighbors = set()
    arcs = layout_data.get("arcs", {})
    if direction == 'up':
        for arc_id, arc_data in arcs.items():
            if arc_data['target'] == vertex_id:
                neighbors.add(arc_data['source'])
    elif direction == 'down':
        for arc_id, arc_data in arcs.items():
            if arc_data['source'] == vertex_id:
                neighbors.add(arc_data['target'])
    return list(neighbors)


def get_arcs_between(u: str, v: str, layout_data: Dict) -> List[Dict]:
    """Gets arcs connecting vertex u and vertex v directly."""
    connecting_arcs = []
    for arc_id, arc_data in layout_data.get("arcs", {}).items():
        if (arc_data['source'] == u and arc_data['target'] == v) or \
                (arc_data['source'] == v and arc_data['target'] == u):
            connecting_arcs.append(arc_data)
    return connecting_arcs


def get_arcs_between_layers(layer_idx1: int, layer_idx2: int, layout_data: Dict) -> List[Dict]:
    """Gets all arcs connecting nodes between two specific layers."""
    arcs_between = []
    vertices = layout_data.get("vertices", {})
    layering = layout_data.get("layering", [])

    if layer_idx1 < 0 or layer_idx1 >= len(layering) or \
            layer_idx2 < 0 or layer_idx2 >= len(layering):
        return []

    layer1_nodes = set(layering[layer_idx1])
    layer2_nodes = set(layering[layer_idx2])

    for arc_id, arc_data in layout_data.get("arcs", {}).items():
        src = arc_data['source']
        tgt = arc_data['target']
        if (src in layer1_nodes and tgt in layer2_nodes) or \
                (src in layer2_nodes and tgt in layer1_nodes):
            if src in layer1_nodes and tgt in layer2_nodes:
                arcs_between.append(arc_data)
            elif src in layer2_nodes and tgt in layer1_nodes:
                arcs_between.append(arc_data)

    return arcs_between


def is_incident_to_inner_segment(ocpn_layout: Dict, vertex_id: str) -> bool:
    """Checks if a vertex is a dummy node connected to another dummy node above it."""
    vertices = ocpn_layout.get("vertices", {})
    v = vertices.get(vertex_id)
    if v and v.get("type") == DUMMY_TYPE:
        upper_neighbor_id = v.get("upper_neighbor")
        if upper_neighbor_id:
            upper_neighbor = vertices.get(upper_neighbor_id)
            if upper_neighbor and upper_neighbor.get("type") == DUMMY_TYPE:
                return True
    return False


# --- Core Sugiyama Steps ---

def _build_initial_layout(ocpn: Dict[str, Any], parameters: Dict) -> Dict:
    object_types = sorted(ocpn.get("object_types", []))
    layout_data = {
        "vertices": {},
        "arcs": {},
        "original_arcs": {},
        "object_types": object_types,
        "config": parameters
    }
    vertex_map = {}

    for ot in object_types:
        net_data = ocpn.get("petri_nets", {}).get(ot)
        if not net_data:
            print(f"Warning: No Petri net found for object type '{ot}', skipping.")
            continue
        net, im, fm = net_data
        # Create places
        for place in net.places:
            place_id = str(place)
            if (PLACE_TYPE, place_id) not in vertex_map:
                layout_id = f"place_{ot}_{place.name}_{uuid.uuid4()}"
                vertex_map[(PLACE_TYPE, place_id)] = layout_id
                is_source = place in im
                is_sink = place in fm
                layout_data["vertices"][layout_id] = {
                    "id": layout_id,
                    "original_id": place_id,
                    "name": place.name or f"p_{ot}",
                    "label": place.name or f"p_{ot}",
                    "objectType": ot,
                    "type": PLACE_TYPE,
                    "layer": -1,
                    "pos": -1,
                    "x": None,
                    "y": None,
                    "is_source": is_source,
                    "is_sink": is_sink,
                    "root": layout_id,
                    "align": layout_id,
                    "sink": layout_id,
                    "shift": float('inf'),
                    "upper_neighbor": None,
                    "lower_neighbor": None,
                }

        # Create transitions
        for trans in net.transitions:
            trans_id = str(trans)
            layout_id = None
            if trans.label:
                if (TRANSITION_TYPE, trans.label) in vertex_map:
                    layout_id = vertex_map[(TRANSITION_TYPE, trans.label)]
                    layout_data["vertices"][layout_id].setdefault("adjacentObjectTypes", set()).add(ot)
                else:
                    layout_id = f"trans_{trans.label}_{uuid.uuid4()}"
                    vertex_map[(TRANSITION_TYPE, trans.label)] = layout_id
                    layout_data["vertices"][layout_id] = {
                        "id": layout_id,
                        "original_id": trans_id,
                        "name": trans.label,
                        "label": trans.label,
                        "objectType": None,
                        "adjacentObjectTypes": {ot},
                        "type": TRANSITION_TYPE,
                        "silent": False,
                        "layer": -1,
                        "pos": -1,
                        "x": None,
                        "y": None,
                        "root": layout_id,
                        "align": layout_id,
                        "sink": layout_id,
                        "shift": float('inf'),
                        "upper_neighbor": None,
                        "lower_neighbor": None,
                    }
            else:
                layout_id = f"silent_{ot}_{trans.name}_{uuid.uuid4()}"
                vertex_map[(TRANSITION_TYPE, trans_id)] = layout_id
                layout_data["vertices"][layout_id] = {
                    "id": layout_id,
                    "original_id": trans_id,
                    "name": trans.name or "silent",
                    "label": None,
                    "objectType": ot,
                    "adjacentObjectTypes": {ot},
                    "type": TRANSITION_TYPE,
                    "silent": True,
                    "layer": -1,
                    "pos": -1,
                    "x": None,
                    "y": None,
                    "root": layout_id,
                    "align": layout_id,
                    "sink": layout_id,
                    "shift": float('inf'),
                    "upper_neighbor": None,
                    "lower_neighbor": None,
                }

        # Create arcs
        for arc in net.arcs:
            source_node = arc.source
            target_node = arc.target
            source_id_orig = str(source_node)
            target_id_orig = str(target_node)

            source_type = PLACE_TYPE if isinstance(source_node, PetriNet.Place) else TRANSITION_TYPE
            target_type = PLACE_TYPE if isinstance(target_node, PetriNet.Place) else TRANSITION_TYPE

            if source_type == TRANSITION_TYPE and source_node.label:
                source_layout_id = vertex_map.get((source_type, source_node.label))
            else:
                source_layout_id = vertex_map.get((source_type, source_id_orig))

            if target_type == TRANSITION_TYPE and target_node.label:
                target_layout_id = vertex_map.get((target_type, target_node.label))
            else:
                target_layout_id = vertex_map.get((target_type, target_id_orig))

            if source_layout_id and target_layout_id:
                arc_exists = False
                existing_arc_id = None
                for arc_id_check, arc_data_check in layout_data["arcs"].items():
                    if (arc_data_check['source'] == source_layout_id and arc_data_check['target'] == target_layout_id) or \
                       (arc_data_check['source'] == target_layout_id and arc_data_check['target'] == source_layout_id):
                        arc_exists = True
                        existing_arc_id = arc_id_check
                        break

                if not arc_exists:
                    arc_id = generate_arc_id(source_layout_id, target_layout_id)
                    is_variable = False
                    if ot in ocpn.get("double_arcs_on_activity", {}):
                        act_label = None
                        if source_type == TRANSITION_TYPE and source_node.label:
                            act_label = source_node.label
                        elif target_type == TRANSITION_TYPE and target_node.label:
                            act_label = target_node.label
                        if act_label and ocpn["double_arcs_on_activity"][ot].get(act_label, False):
                            is_variable = True

                    arc_data = {
                        "id": arc_id,
                        "source": source_layout_id,   # keep ORIGINAL semantics
                        "target": target_layout_id,   # keep ORIGINAL semantics
                        "objectType": ot,
                        "original": True,
                        "reversed": False,            # will be set True if reversed for DAG
                        "dummy_nodes": [],
                        "minLayer": -1,
                        "maxLayer": -1,
                        "type1": False,
                        "weight": arc.weight if hasattr(arc, 'weight') else 1,
                        "variable": is_variable,
                    }
                    layout_data["arcs"][arc_id] = arc_data
                    layout_data["original_arcs"][arc_id] = copy.deepcopy(arc_data)
                else:
                    if source_type == TRANSITION_TYPE:
                        layout_data["vertices"][source_layout_id].setdefault("adjacentObjectTypes", set()).add(ot)
                    if target_type == TRANSITION_TYPE:
                        layout_data["vertices"][target_layout_id].setdefault("adjacentObjectTypes", set()).add(ot)
                    pass
            else:
                print(f"Warning: Could not find layout nodes for arc {source_id_orig} -> {target_id_orig} in OT {ot}")

    all_activity_nodes = {v['name'] for v in layout_data['vertices'].values() if
                          v['type'] == TRANSITION_TYPE and not v['silent']}
    for act in ocpn.get("activities", set()):
        if act not in all_activity_nodes:
            layout_id = f"trans_{act}_unconnected_{uuid.uuid4()}"
            layout_data["vertices"][layout_id] = {
                "id": layout_id, "original_id": None, "name": act, "label": act,
                "objectType": None, "adjacentObjectTypes": set(), "type": TRANSITION_TYPE,
                "silent": False, "layer": -1, "pos": -1, "x": None, "y": None,
                "root": layout_id, "align": layout_id, "sink": layout_id, "shift": float('inf'),
                "upper_neighbor": None, "lower_neighbor": None,
            }
            print(f"Warning: Activity '{act}' not found in Petri net transitions, adding as unconnected node.")

    return layout_data


# --- Graph representation for algorithms ---
class OCPNGraph:
    """Helper class to represent the graph for cycle breaking and layer assignment."""

    def __init__(self, layout_data: Dict):
        self.nodes = list(layout_data["vertices"].keys())
        self.arcs_data = layout_data["arcs"]
        self.adj = {n: [] for n in self.nodes}
        self.rev_adj = {n: [] for n in self.nodes}
        self.arc_list = []

        for arc_id, arc in self.arcs_data.items():
            src, tgt = arc['source'], arc['target']
            if src == tgt:
                continue
            if src in self.nodes and tgt in self.nodes:
                self.adj[src].append(tgt)
                self.rev_adj[tgt].append(src)
                self.arc_list.append({'source': src, 'target': tgt, 'id': arc_id})

    def get_out_degree(self, node: str) -> int:
        return len(self.adj.get(node, []))

    def get_in_degree(self, node: str) -> int:
        return len(self.rev_adj.get(node, []))

    def remove_node(self, node: str):
        if node not in self.nodes:
            return
        successors = self.adj.pop(node, [])
        predecessors = self.rev_adj.pop(node, [])
        for succ in successors:
            if succ in self.rev_adj:
                self.rev_adj[succ].remove(node)
        for pred in predecessors:
            if pred in self.adj:
                self.adj[pred].remove(node)
        self.nodes.remove(node)
        self.arc_list = [a for a in self.arc_list if a['source'] != node and a['target'] != node]

    def remove_nodes(self, nodes_to_remove: List[str]):
        for node in nodes_to_remove:
            self.remove_node(node)

    def get_sink(self) -> Optional[str]:
        for node in self.nodes:
            if self.get_out_degree(node) == 0:
                return node
        return None

    def get_source(self) -> Optional[str]:
        for node in self.nodes:
            if self.get_in_degree(node) == 0:
                return node
        return None

    def get_max_out_degree_node(self) -> Optional[str]:
        max_degree = -1
        max_node = None
        for node in self.nodes:
            degree = self.get_out_degree(node) - self.get_in_degree(node)
            if degree > max_degree:
                max_degree = degree
                max_node = node
        return max_node


def _modified_greedy_fas(graph: OCPNGraph, sources: List[str], sinks: List[str]) -> List[str]:
    """Computes Feedback Arc Set using greedy algorithm considering sources/sinks."""
    s1 = list(sources)
    s2 = list(sinks)

    s1 = [n for n in s1 if n in graph.nodes]
    s2 = [n for n in s2 if n in graph.nodes]

    if s1:
        s1.sort(key=lambda n: graph.get_out_degree(n) - graph.get_in_degree(n), reverse=True)
        graph.remove_nodes(s1)

    if s2:
        s2.sort(key=lambda n: graph.get_out_degree(n) - graph.get_in_degree(n), reverse=True)
        graph.remove_nodes(s2)

    temp_s1 = []
    temp_s2 = []
    while graph.nodes:
        sink = graph.get_sink()
        while sink:
            temp_s2.insert(0, sink)
            graph.remove_node(sink)
            sink = graph.get_sink()

        source = graph.get_source()
        while source:
            temp_s1.append(source)
            graph.remove_node(source)
            source = graph.get_source()

        if graph.nodes:
            node = graph.get_max_out_degree_node()
            if node:
                temp_s1.append(node)
                graph.remove_node(node)
            else:
                print("Warning: No node found in FAS loop, graph might be empty or disconnected unexpectedly.")
                break

    return s1 + temp_s1 + temp_s2 + s2


def _reverse_cycles(layout_data: Dict, config: Dict) -> int:
    """Marks arcs to reverse (for DAG) without swapping endpoints."""
    print("Step 1: Cycle Breaking...")
    graph = OCPNGraph(layout_data)

    potential_sources = [vid for vid, vdata in layout_data["vertices"].items() if vdata.get("is_source")]
    potential_sinks = [vid for vid, vdata in layout_data["vertices"].items() if vdata.get("is_sink")]
    sources = config.get("sources", potential_sources)
    sinks = config.get("sinks", potential_sinks)

    sources = [s for s in sources if s in layout_data["vertices"]]
    sinks = [s for s in sinks if s in layout_data["vertices"]]

    fas_order = _modified_greedy_fas(graph, sources, sinks)
    fas_indices = {node_id: index for index, node_id in enumerate(fas_order)}

    reversed_count = 0
    for arc_id, arc in layout_data["arcs"].items():
        src = arc['source']
        tgt = arc['target']
        src_index = fas_indices.get(src)
        tgt_index = fas_indices.get(tgt)

        if src_index is not None and tgt_index is not None:
            if src_index > tgt_index:
                # Do NOT swap. Just mark as reversed for layout purposes.
                arc['reversed'] = True
                reversed_count += 1
            else:
                arc['reversed'] = False
        else:
            arc['reversed'] = False

    print(f"  Marked {reversed_count} arcs as reversed for DAG.")
    return reversed_count


def _assign_layers(layout_data: Dict) -> bool:
    """Assigns layers to vertices using LP to minimize edge length."""
    print("Step 2: Layer Assignment (using LP)...")
    vertices = layout_data["vertices"]
    arcs = layout_data["arcs"]
    node_list = list(vertices.keys())
    node_to_index = {node_id: i for i, node_id in enumerate(node_list)}
    num_vars = len(node_list)

    # Objective: Minimize sum(weight * (layer[target] - layer[source]))
    c = np.zeros(num_vars)
    active_arcs = []

    for arc_id, arc in arcs.items():
        src_orig, tgt_orig = arc['source'], arc['target']
        weight = arc.get('weight', 1)

        # Use the 'reversed' flag to determine DAG direction
        if arc['reversed']:
            src, tgt = tgt_orig, src_orig
        else:
            src, tgt = src_orig, tgt_orig

        if src in node_to_index and tgt in node_to_index:
            src_idx = node_to_index[src]
            tgt_idx = node_to_index[tgt]
            c[tgt_idx] += weight
            c[src_idx] -= weight
            active_arcs.append({'source': src, 'target': tgt, 'id': arc_id})

    num_constraints = len(active_arcs)
    Aub = np.zeros((num_constraints, num_vars))
    bub = np.full(num_constraints, -1.0)

    for i, arc in enumerate(active_arcs):
        src_idx = node_to_index[arc['source']]
        tgt_idx = node_to_index[arc['target']]
        Aub[i, src_idx] = 1.0
        Aub[i, tgt_idx] = -1.0

    num_positivity_constraints = num_vars
    Aub_pos = np.zeros((num_positivity_constraints, num_vars))
    bub_pos = np.zeros(num_positivity_constraints)
    for i in range(num_vars):
        Aub_pos[i, i] = -1.0

    Aub_combined = np.vstack((Aub, Aub_pos))
    bub_combined = np.concatenate((bub, bub_pos))

    Aeq = None
    beq = None

    try:
        print(f"  Setting up LP with {num_vars} variables and {Aub_combined.shape[0]} constraints.")
        sol = lp_solver.apply(c, Aub_combined, bub_combined, Aeq, beq,
                              parameters={"verbose": False}, variant=lp_solver.SCIPY)

        if sol is None:
            print("  LP Solver did not return a solution.")
            return False

        points = lp_solver.get_points_from_sol(sol, variant=lp_solver.SCIPY)

        if points is None or len(points) != num_vars:
            print("  LP Solver failed to extract points or returned incorrect number.")
            return False

        max_layer = 0
        layering_dict = {}
        for i, node_id in enumerate(node_list):
            layer = int(round(points[i]))
            vertices[node_id]['layer'] = layer
            if layer not in layering_dict:
                layering_dict[layer] = []
            layering_dict[layer].append(node_id)
            max_layer = max(max_layer, layer)

        layout_data['layering'] = []
        for i in range(max_layer + 1):
            layout_data['layering'].append(layering_dict.get(i, []))

        print(f"  Assigned nodes to {max_layer + 1} layers.")
        return True

    except Exception as e:
        print(f"  Error during LP solving or layer assignment: {e}")
        import traceback
        traceback.print_exc()
        return False


def _insert_dummy_vertices(layout_data: Dict):
    """Inserts dummy nodes for arcs spanning multiple layers."""
    print("Step 3: Dummy Vertex Insertion...")
    vertices = layout_data["vertices"]
    arcs = layout_data["arcs"]
    layering = layout_data["layering"]
    arcs_to_add = {}
    arcs_to_remove = []
    dummy_count = 0

    original_arc_ids = list(arcs.keys())

    for arc_id in original_arc_ids:
        arc = arcs[arc_id]
        src_id_orig, tgt_id_orig = arc['source'], arc['target']

        # Use DAG direction when computing span
        if arc['reversed']:
            src_id, tgt_id = tgt_id_orig, src_id_orig
        else:
            src_id, tgt_id = src_id_orig, tgt_id_orig

        if src_id not in vertices or tgt_id not in vertices:
            print(f"Warning: Skipping dummy insertion for arc {arc_id} - source or target not found.")
            continue

        src_node = vertices[src_id]
        tgt_node = vertices[tgt_id]

        src_layer = src_node.get('layer', -1)
        tgt_layer = tgt_node.get('layer', -1)

        if src_layer == -1 or tgt_layer == -1:
            print(f"Warning: Skipping dummy insertion for arc {arc_id} ({src_id} -> {tgt_id}) - layers not assigned.")
            continue

        arc['minLayer'] = min(src_layer, tgt_layer)
        arc['maxLayer'] = max(src_layer, tgt_layer)

        slack = tgt_layer - src_layer

        if slack > 1:
            arcs_to_remove.append(arc_id)
            dummies_in_path = []

            ot = arc.get("objectType")
            if vertices[src_id_orig]["type"] == PLACE_TYPE:
                ot = vertices[src_id_orig]["objectType"]
            elif vertices[tgt_id_orig]["type"] == PLACE_TYPE:
                ot = vertices[tgt_id_orig]["objectType"]

            src_pos = -1
            tgt_pos = -1
            if src_layer < len(layering) and src_id in layering[src_layer]:
                src_pos = layering[src_layer].index(src_id)
            if tgt_layer < len(layering) and tgt_id in layering[tgt_layer]:
                tgt_pos = layering[tgt_layer].index(tgt_id)

            median_pos = len(layering[src_layer + 1]) // 2
            if src_pos != -1 and tgt_pos != -1:
                median_pos = (src_pos + tgt_pos) // 2

            last_node_id = src_id
            for i in range(1, slack):
                dummy_layer_idx = src_layer + i
                dummy_id = generate_dummy_id()
                dummy_count += 1
                dummies_in_path.append(dummy_id)

                dummy_node = {
                    "id": dummy_id,
                    "original_id": None,
                    "name": f"dummy_{dummy_count}",
                    "label": None,
                    "objectType": ot,
                    "type": DUMMY_TYPE,
                    "layer": dummy_layer_idx,
                    "pos": -1,
                    "x": None,
                    "y": None,
                    "dummy_of_arc": arc_id,
                    "root": dummy_id,
                    "align": dummy_id,
                    "sink": dummy_id,
                    "shift": float('inf'),
                    "upper_neighbor": last_node_id,
                    "lower_neighbor": None,
                }
                vertices[dummy_id] = dummy_node

                if i > 1:
                    vertices[dummies_in_path[-2]]["lower_neighbor"] = dummy_id

                if dummy_layer_idx < len(layering):
                    insert_pos = min(median_pos, len(layering[dummy_layer_idx]))
                    layering[dummy_layer_idx].insert(insert_pos, dummy_id)
                else:
                    print(f"Warning: Dummy layer index {dummy_layer_idx} out of bounds.")
                    layering.append([dummy_id])

                segment_arc_id = generate_arc_id(last_node_id, dummy_id)
                arcs_to_add[segment_arc_id] = {
                    "id": segment_arc_id,
                    "source": last_node_id,
                    "target": dummy_id,
                    "objectType": ot,
                    "original": False,
                    "reversed": False,
                    "dummy_nodes": [],
                    "minLayer": dummy_layer_idx - 1,
                    "maxLayer": dummy_layer_idx,
                    "type1": False,
                    "weight": arc.get('weight', 1),
                    "variable": arc.get('variable', False),
                    "original_arc_id": arc_id
                }

                last_node_id = dummy_id

            segment_arc_id = generate_arc_id(last_node_id, tgt_id)
            arcs_to_add[segment_arc_id] = {
                "id": segment_arc_id,
                "source": last_node_id,
                "target": tgt_id,
                "objectType": ot,
                "original": False,
                "reversed": False,
                "dummy_nodes": [],
                "minLayer": tgt_layer - 1,
                "maxLayer": tgt_layer,
                "type1": False,
                "weight": arc.get('weight', 1),
                "variable": arc.get('variable', False),
                "original_arc_id": arc_id
            }
            vertices[last_node_id]["lower_neighbor"] = tgt_id

            if arc_id in layout_data["original_arcs"]:
                layout_data["original_arcs"][arc_id]["dummy_nodes"] = dummies_in_path

            arcs[arc_id]['dummy_nodes'] = dummies_in_path

    for arc_id in arcs_to_remove:
        del arcs[arc_id]
    arcs.update(arcs_to_add)

    print(f"  Inserted {dummy_count} dummy vertices.")


# --- Vertex Ordering (Barycenter Heuristic) ---

def _compute_barycenter(ocpn_layout: Dict, vertex_id: str, layer_idx: int, fixed_layer_idx: int, config: Dict) -> float:
    """Computes the barycenter value for a vertex based on neighbors in the fixed layer."""
    vertices = ocpn_layout["vertices"]
    layering = ocpn_layout["layering"]

    if fixed_layer_idx < 0 or fixed_layer_idx >= len(layering):
        return 0.0

    fixed_layer = layering[fixed_layer_idx]
    fixed_layer_positions = {node_id: pos for pos, node_id in enumerate(fixed_layer)}

    neighbors_in_fixed = []
    arcs = ocpn_layout.get("arcs", {})
    for arc_id, arc in arcs.items():
        src_orig, tgt_orig = arc['source'], arc['target']
        other_node = None
        if src_orig == vertex_id and tgt_orig in fixed_layer_positions:
            other_node = tgt_orig
        elif tgt_orig == vertex_id and src_orig in fixed_layer_positions:
            other_node = src_orig

        if other_node:
            other_node_layer = vertices.get(other_node, {}).get('layer', -1)
            if other_node_layer == fixed_layer_idx:
                neighbors_in_fixed.append(other_node)

    if not neighbors_in_fixed:
        return -1.0

    barycenter_sum = 0.0
    for neighbor in neighbors_in_fixed:
        barycenter_sum += fixed_layer_positions[neighbor]

    avg_barycenter = barycenter_sum / len(neighbors_in_fixed)

    if vertices[vertex_id]["type"] == PLACE_TYPE and config.get("object_attraction", 0) > 0:
        object_attraction = config.get("object_attraction", 0.5)
        obj_attr_min = config.get("object_attraction_range_min", 1)
        obj_attr_max = config.get("object_attraction_range_max", 1)
        current_ot = vertices[vertex_id]["objectType"]

        object_neighbor_positions = []
        direction = 1 if fixed_layer_idx > layer_idx else -1

        for i in range(obj_attr_min, obj_attr_max + 1):
            target_layer_idx = layer_idx + direction * (2 * i)
            if 0 <= target_layer_idx < len(layering):
                target_layer = layering[target_layer_idx]
                for neighbor_pos, node_id in enumerate(target_layer):
                    if vertices[node_id]["type"] == PLACE_TYPE and vertices[node_id]["objectType"] == current_ot:
                        object_neighbor_positions.append(neighbor_pos)

        if object_neighbor_positions:
            avg_object_pos = sum(object_neighbor_positions) / len(object_neighbor_positions)
            return (1.0 - object_attraction) * avg_barycenter + object_attraction * avg_object_pos
        else:
            return avg_barycenter
    elif vertices[vertex_id]["type"] == DUMMY_TYPE:
        if neighbors_in_fixed:
            return fixed_layer_positions[neighbors_in_fixed[0]]
        else:
            return -1.0
    else:
        return avg_barycenter


def _barycenter_ordering_pass(ocpn_layout: Dict, current_layering: List[List[str]], fixed_layer_idx: int,
                              sweep_down: bool, config: Dict) -> List[List[str]]:
    new_layering = copy.deepcopy(current_layering)
    layer_to_order_idx = fixed_layer_idx + 1 if sweep_down else fixed_layer_idx - 1

    if layer_to_order_idx < 0 or layer_to_order_idx >= len(new_layering):
        return new_layering

    layer_to_order = new_layering[layer_to_order_idx]
    if not layer_to_order:
        return new_layering

    barycenters = {}
    for node_id in layer_to_order:
        bc_val = _compute_barycenter(ocpn_layout, node_id, layer_to_order_idx, fixed_layer_idx, config)
        barycenters[node_id] = bc_val

    adjusted_barycenters = {}
    last_valid_bc = 0.0
    for i, node_id in enumerate(layer_to_order):
        bc = barycenters[node_id]
        if bc == -1.0:
            adjusted_bc = last_valid_bc + i * 1e-6
            adjusted_barycenters[node_id] = adjusted_bc
        else:
            adjusted_barycenters[node_id] = bc
            last_valid_bc = bc

    original_indices = {node_id: i for i, node_id in enumerate(layer_to_order)}

    def sort_key(node_id):
        return (adjusted_barycenters[node_id], original_indices[node_id])

    ordered_layer = sorted(layer_to_order, key=sort_key)

    new_layering[layer_to_order_idx] = ordered_layer
    return new_layering


def _count_crossings(ocpn_layout: Dict, layering: List[List[str]]) -> int:
    crossings = 0
    vertices = ocpn_layout["vertices"]
    arcs_data = ocpn_layout.get("arcs", {})

    node_positions = {}
    for r, layer in enumerate(layering):
        for c, node_id in enumerate(layer):
            node_positions[node_id] = (r, c)

    for r in range(len(layering) - 1):
        layer1 = layering[r]
        layer2 = layering[r + 1]

        arcs_between = []
        for arc_id, arc in arcs_data.items():
            src, tgt = arc['source'], arc['target']
            pos_src = node_positions.get(src)
            pos_tgt = node_positions.get(tgt)
            if pos_src and pos_tgt:
                if (pos_src[0] == r and pos_tgt[0] == r + 1):
                    arcs_between.append(
                        {'source': src, 'target': tgt, 'source_pos': pos_src[1], 'target_pos': pos_tgt[1]})
                elif (pos_src[0] == r + 1 and pos_tgt[0] == r):
                    arcs_between.append(
                        {'source': tgt, 'target': src, 'source_pos': pos_tgt[1], 'target_pos': pos_src[1]})

        for i in range(len(arcs_between)):
            for j in range(i + 1, len(arcs_between)):
                arc1 = arcs_between[i]
                arc2 = arcs_between[j]
                if (arc1['source_pos'] < arc2['source_pos'] and arc1['target_pos'] > arc2['target_pos']) or \
                   (arc1['source_pos'] > arc2['source_pos'] and arc1['target_pos'] < arc2['target_pos']):
                    crossings += 1

    return crossings


def _order_vertices(layout_data: Dict, config: Dict):
    """Orders vertices within layers using the barycenter method."""
    print("Step 4: Vertex Ordering (Barycenter Heuristic)...")

    max_iterations = config.get("max_barycenter_iterations", 24)

    current_layering = layout_data["layering"]
    best_layering = copy.deepcopy(current_layering)
    best_score = _count_crossings(layout_data, best_layering)
    print(f"  Initial crossing score: {best_score}")

    no_improvement_counter = 0
    iter_count = 0

    while no_improvement_counter < max_iterations:
        iter_count += 1
        layering_before_sweep = copy.deepcopy(current_layering)

        for i in range(len(current_layering) - 1):
            current_layering = _barycenter_ordering_pass(layout_data, current_layering, fixed_layer_idx=i,
                                                         sweep_down=True, config=config)

        for i in range(len(current_layering) - 1, 0, -1):
            current_layering = _barycenter_ordering_pass(layout_data, current_layering, fixed_layer_idx=i,
                                                         sweep_down=False, config=config)

        current_score = _count_crossings(layout_data, current_layering)

        if current_score < best_score:
            best_score = current_score
            best_layering = copy.deepcopy(current_layering)
            no_improvement_counter = 0
            print(f"  Iteration {iter_count}: Improved score to {best_score}")
        else:
            no_improvement_counter += 1
            print(f"  Iteration {iter_count}: Score {current_score} (no improvement)")

        if current_layering == layering_before_sweep:
            print(f"  Converged after {iter_count} iterations.")
            break

    print(f"  Final crossing score: {best_score} after {iter_count} iterations.")
    layout_data["layering"] = best_layering

    for r, layer in enumerate(best_layering):
        for c, node_id in enumerate(layer):
            layout_data["vertices"][node_id]['pos'] = c


# --- Vertex Positioning (Brandes & Köpf Heuristic Variant) ---

def _mark_type1_conflicts(layout_data: Dict):
    vertices = layout_data["vertices"]
    arcs = layout_data["arcs"]
    layering = layout_data["layering"]

    for arc_id, arc in arcs.items():
        arc['type1'] = False

    for i in range(len(layering) - 2):
        layer_i = layering[i]
        layer_i1 = layering[i + 1]

        pos_i1 = {node_id: pos for pos, node_id in enumerate(layer_i1)}
        pos_i = {node_id: pos for pos, node_id in enumerate(layer_i)}

        inner_segments = []
        for k, v_i1 in enumerate(layer_i1):
            node_i1_data = vertices[v_i1]
            if node_i1_data['type'] == DUMMY_TYPE:
                upper_neighbor_id = node_i1_data.get("upper_neighbor")
                if upper_neighbor_id and upper_neighbor_id in pos_i:
                    upper_neighbor_data = vertices[upper_neighbor_id]
                    if upper_neighbor_data['type'] == DUMMY_TYPE:
                        inner_segments.append((pos_i[upper_neighbor_id], k))

        if not inner_segments:
            continue

        inner_segments.sort()

        k0 = 0
        current_inner_idx = 0
        for l1, v_i1 in enumerate(layer_i1):
            k1 = len(layer_i) - 1
            is_lower_end_of_inner = False

            while current_inner_idx < len(inner_segments) and inner_segments[current_inner_idx][1] <= l1:
                k1 = inner_segments[current_inner_idx][0]
                is_lower_end_of_inner = (inner_segments[current_inner_idx][1] == l1)
                current_inner_idx += 1

            if not is_lower_end_of_inner:
                upper_neighbors_i = []
                for arc_id, arc_data in arcs.items():
                    if arc_data['target'] == v_i1 and arc_data['source'] in pos_i:
                        upper_neighbors_i.append(arc_data['source'])

                for u_i in upper_neighbors_i:
                    k = pos_i[u_i]
                    if k < k0 or k > k1:
                        connecting_arcs = get_arcs_between(u_i, v_i1, layout_data)
                        for conn_arc in connecting_arcs:
                            is_seg_inner = (vertices[u_i]['type'] == DUMMY_TYPE and vertices[v_i1]['type'] == DUMMY_TYPE)
                            if not is_seg_inner:
                                conn_arc['type1'] = True

            if is_lower_end_of_inner:
                k0 = k1


def _vertical_alignment(ocpn_layout: Dict, current_layering: List[List[str]], pos: Dict[str, int], sweep_down: bool) -> \
Tuple[Dict[str, str], Dict[str, str]]:
    root = {vid: vid for vid in ocpn_layout["vertices"]}
    align = {vid: vid for vid in ocpn_layout["vertices"]}
    vertices = ocpn_layout["vertices"]

    layer_indices = range(len(current_layering)) if sweep_down else range(len(current_layering) - 1, -1, -1)

    for i in layer_indices:
        fixed_layer_idx = i - 1 if sweep_down else i + 1
        if fixed_layer_idx < 0 or fixed_layer_idx >= len(current_layering):
            continue

        layer_k = current_layering[i]
        r = -1

        for vk in layer_k:
            neighbors = []
            neighbor_dir = 'up' if sweep_down else 'down'

            arcs_data = ocpn_layout.get("arcs", {})
            fixed_layer_nodes = set(current_layering[fixed_layer_idx])

            for arc_id, arc in arcs_data.items():
                src, tgt = arc['source'], arc['target']
                if sweep_down:
                    if tgt == vk and src in fixed_layer_nodes:
                        neighbors.append(src)
                else:
                    if src == vk and tgt in fixed_layer_nodes:
                        neighbors.append(tgt)

            if not neighbors:
                continue

            neighbors.sort(key=lambda n: pos[n])

            m_low = math.floor((len(neighbors) - 1) / 2)
            m_high = math.ceil((len(neighbors) - 1) / 2)
            median_indices = list(set([m_low, m_high]))

            for m_idx in median_indices:
                median_neighbor = neighbors[m_idx]

                if align[vk] == vk:
                    connecting_arcs = get_arcs_between(median_neighbor, vk, ocpn_layout)
                    is_marked = any(a.get('type1', False) for a in connecting_arcs)

                    if not is_marked and pos[median_neighbor] > r:
                        align[median_neighbor] = vk
                        root[vk] = root[median_neighbor]
                        align[vk] = root[vk]
                        r = pos[median_neighbor]

    return root, align


def _place_block(ocpn_layout: Dict, layering: List[List[str]], v: str, x: Dict[str, Optional[float]],
                 pos: Dict[str, int], roots: Dict[str, str], sink: Dict[str, str], shift: Dict[str, float],
                 aligns: Dict[str, str], config: Dict):
    if x.get(v) is None:
        x[v] = 0.0
        w = v
        while True:
            w_pos = pos[w]
            w_layer_idx = -1
            for idx, l in enumerate(layering):
                if w in l:
                    w_layer_idx = idx
                    break

            if w_pos > 0 and w_layer_idx != -1:
                predecessor = layering[w_layer_idx][w_pos - 1]
                u = roots[predecessor]
                _place_block(ocpn_layout, layering, u, x, pos, roots, sink, shift, aligns, config)

                if sink[v] == v:
                    sink[v] = sink[u]

                place_dim = config.get("place_radius", 10) * 2
                trans_h = config.get("transition_height", 20)
                trans_w = config.get("transition_width", 50)
                silent_w = config.get("silent_transition_width", 10)
                max_trans_dim = max(trans_h, trans_w, silent_w)

                delta = config.get("vertex_sep", 20) + max(place_dim, max_trans_dim)

                if sink[v] != sink[u]:
                    current_shift = (x.get(v, 0.0) or 0.0) - (x.get(u, 0.0) or 0.0) - delta
                    shift[sink[u]] = min(shift.get(sink[u], float('inf')), current_shift)
                else:
                    new_x = (x.get(u, 0.0) or 0.0) + delta
                    x[v] = max(x.get(v, 0.0) or 0.0, new_x)

            w = aligns[w]
            if w == v:
                break


def _horizontal_compaction(ocpn_layout: Dict, layering: List[List[str]], roots: Dict[str, str], aligns: Dict[str, str],
                           pos: Dict[str, int], config: Dict) -> Tuple[Dict[str, Optional[float]], float]:
    x: Dict[str, Optional[float]] = {vid: None for vid in ocpn_layout["vertices"]}
    sink: Dict[str, str] = {vid: vid for vid in ocpn_layout["vertices"]}
    shift: Dict[str, float] = {vid: float('inf') for vid in ocpn_layout["vertices"]}

    for layer in layering:
        for v in layer:
            if roots[v] == v:
                _place_block(ocpn_layout, layering, v, x, pos, roots, sink, shift, aligns, config)

    abs_x: Dict[str, Optional[float]] = {}
    max_coord = 0.0
    min_coord = float('inf')

    for layer in layering:
        for v in layer:
            root_v = roots[v]
            if x.get(root_v) is not None:
                current_x = x[root_v]
                sink_root_v = sink[root_v]
                if shift.get(sink_root_v, float('inf')) != float('inf'):
                    current_x += shift[sink_root_v]
                abs_x[v] = current_x
                max_coord = max(max_coord, current_x)
                min_coord = min(min_coord, current_x)
            else:
                abs_x[v] = None

    border = config.get("border_padding", 20)
    shift_val = border - min_coord if min_coord != float('inf') else border
    final_x = {}
    final_max = 0.0
    for v, val in abs_x.items():
        if val is not None:
            final_x[v] = val + shift_val
            final_max = max(final_max, final_x[v])
        else:
            final_x[v] = None

    final_max += border

    return final_x, final_max


def _position_vertices(layout_data: Dict, config: Dict):
    """Assigns final X/Y coordinates using 4-pass heuristic."""
    print("Step 5: Vertex Positioning...")

    _mark_type1_conflicts(layout_data)

    layouts = []

    original_layering = copy.deepcopy(layout_data["layering"])

    for vert_dir in [0, 1]:  # 0: Down, 1: Up
        for horz_dir in [0, 1]:  # 0: Left->Right, 1: Right->Left

            print(f"  Running positioning pass (vert_dir={vert_dir}, horz_dir={horz_dir})...")
            current_layering = copy.deepcopy(original_layering)

            if vert_dir == 1:
                current_layering.reverse()
            if horz_dir == 1:
                for layer in current_layering:
                    layer.reverse()

            pos = {}
            for r, layer in enumerate(current_layering):
                for c, node_id in enumerate(layer):
                    pos[node_id] = c

            roots, aligns = _vertical_alignment(layout_data, current_layering, pos, sweep_down=(vert_dir == 0))

            coords, max_coord = _horizontal_compaction(layout_data, current_layering, roots, aligns, pos, config)

            if horz_dir == 1:
                final_coords = {}
                for v, c in coords.items():
                    if c is not None:
                        final_coords[v] = max_coord - c
                    else:
                        final_coords[v] = None
                layouts.append(final_coords)
            else:
                layouts.append(coords)

    final_coords_x = {}
    final_coords_y = {}
    vertices = layout_data["vertices"]
    layering = layout_data["layering"]
    border_padding = config.get("border_padding", 20)
    layer_sep = config.get("layer_sep", 50)
    direction = config.get("rankdir", "TB")

    for v_id in vertices.keys():
        candidate_coords = []
        for layout_pass in layouts:
            coord = layout_pass.get(v_id)
            if coord is not None:
                candidate_coords.append(coord)

        if candidate_coords:
            candidate_coords.sort()
            if len(candidate_coords) > 0:
                median_idx1 = math.floor((len(candidate_coords) - 1) / 2)
                median_idx2 = math.ceil((len(candidate_coords) - 1) / 2)
                median_coord = (candidate_coords[median_idx1] + candidate_coords[median_idx2]) / 2
                final_coords_x[v_id] = median_coord
        else:
            final_coords_x[v_id] = border_padding

    layer_sizes = []
    max_layer_dim = 0
    for i, layer in enumerate(layering):
        max_dim = 0
        for node_id in layer:
            node = vertices[node_id]
            node_h, node_w = 0, 0
            if node['type'] == PLACE_TYPE:
                r = config.get("place_radius", 10)
                node_h, node_w = r * 2, r * 2
            elif node['type'] == TRANSITION_TYPE:
                node_h = config.get("transition_height", 20)
                if node['silent']:
                    node_w = config.get("silent_transition_width", 10)
                else:
                    node_w = config.get("transition_width", 50)
            elif node['type'] == DUMMY_TYPE:
                node_h, node_w = 1, 1

            if direction == "TB":
                max_dim = max(max_dim, node_h)
            else:
                max_dim = max(max_dim, node_w)
        layer_sizes.append({'layer': i, 'size': max_dim})
        max_layer_dim = max(max_layer_dim, max_dim)

    current_pos = border_padding
    layer_centers = {}
    for i in range(len(layering)):
        layer_size = layer_sizes[i]['size'] if i < len(layer_sizes) else max_layer_dim
        layer_center = current_pos + layer_size / 2.0
        layer_centers[i] = layer_center
        current_pos += layer_size + layer_sep

    max_x, max_y = 0.0, 0.0
    for v_id, node in vertices.items():
        x_coord = final_coords_x.get(v_id, border_padding)
        y_coord = layer_centers.get(node['layer'], border_padding)

        if direction == "TB":
            node['x'] = x_coord
            node['y'] = y_coord
            max_x = max(max_x, x_coord)
            max_y = max(max_y, y_coord)
        else:
            node['x'] = y_coord
            node['y'] = x_coord
            max_x = max(max_x, y_coord)
            max_y = max(max_y, x_coord)

    layout_data["max_x"] = max_x + border_padding
    layout_data["max_y"] = max_y + border_padding

    print(f"  Assigned final coordinates (max_x={max_x:.2f}, max_y={max_y:.2f}).")


# --- Graphviz Drawing ---

def _create_graphviz_digraph(ocpn: Dict, layout_data: Dict, parameters: Dict) -> Digraph:
    """Generates the Graphviz Digraph object from the layout with separate handling for direct edges and waypoint edges."""
    print("Step 6: Generating Graphviz output...")

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, "TB")
    engine = exec_utils.get_param_value(Parameters.ENGINE, parameters, "dot")
    enable_graph_title = exec_utils.get_param_value(Parameters.ENABLE_GRAPH_TITLE, parameters, constants.DEFAULT_ENABLE_GRAPH_TITLES)
    graph_title = exec_utils.get_param_value(Parameters.GRAPH_TITLE, parameters, "Object-Centric Petri Net (Sugiyama Layout)")

    viz = Digraph(
        "ocpn_sugiyama",
        engine=engine,
        graph_attr={
            "bgcolor": bgcolor,
            "rankdir": rankdir,
            "splines": "line",
        },
        node_attr={'shape': 'box'},
    )

    if enable_graph_title:
        viz.attr(label=f'<<FONT POINT-SIZE="20">{graph_title}</FONT>>', labelloc="t")

    vertices = layout_data["vertices"]

    def get_contrasting_color(hex_color: str) -> str:
        hex_color = hex_color.lstrip('#')
        rgb = tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))
        luminance = 0.299 * rgb[0] + 0.587 * rgb[1] + 0.114 * rgb[2]
        return "white" if luminance < 128 else "black"

    # Add Nodes
    for node_id, node in vertices.items():
        if node.get('x') is None or node.get('y') is None:
            print(f"Warning: Node {node_id} has no coordinates, skipping.")
            continue

        pos_x = node['x']
        pos_y = node['y']
        pos_str = f"{pos_x:.2f},{pos_y:.2f}!"

        attrs = {
            "pos": pos_str,
            "label": node['label'] if node['label'] else "",
            "id": node_id
        }

        if node['type'] == PLACE_TYPE:
            ot_color = ot_to_color(node['objectType'])
            font_color = get_contrasting_color(ot_color)
            radius_gv = parameters.get("place_radius", 10) / 72.0 * 1.5
            if node.get("is_source"):
                ot_label = f"{node['objectType']}"
                attrs.update({
                    "shape": "ellipse",
                    "fillcolor": ot_color,
                    "style": "filled",
                    "fontcolor": font_color,
                    "width": f"{radius_gv * 2.5:.2f}",
                    "height": f"{radius_gv * 1.5:.2f}",
                    "fixedsize": "true",
                    "label": f"<<FONT POINT-SIZE='8'>{ot_label}</FONT>>",
                    "labelloc": "c",
                })
                if node.get("is_source") and node.get("is_sink"):
                    attrs["peripheries"] = "2"
            else:
                attrs.update({
                    "shape": "circle",
                    "fillcolor": ot_color,
                    "style": "filled",
                    "fontcolor": font_color,
                    "width": f"{radius_gv * 2:.2f}",
                    "height": f"{radius_gv * 2:.2f}",
                    "fixedsize": "true",
                    "label": "",
                })
                if node.get("is_sink"):
                    attrs["peripheries"] = "2"

        elif node['type'] == TRANSITION_TYPE:
            height_gv = parameters.get("transition_height", 20) / 72.0
            if node.get('silent', False):
                width_gv = parameters.get("silent_transition_width", 10) / 72.0
                attrs.update({
                    "shape": "box",
                    "fillcolor": parameters.get("transition_fill_color", "black"),
                    "style": "filled",
                    "width": f"{width_gv:.2f}",
                    "height": f"{height_gv:.2f}",
                    "fixedsize": "true",
                    "label": "",
                })
            else:
                base_width_gv = parameters.get("transition_width", 40) / 72.0
                width_gv = base_width_gv * 1.5
                font_size = parameters.get("font_size", 10) - 2
                label = node['label']
                max_label_width = 10
                if label and len(label) > max_label_width:
                    wrapped_label = "<BR/>".join([label[i:i+max_label_width] for i in range(0, len(label), max_label_width)])
                    attrs["label"] = f"<{wrapped_label}>"
                else:
                    attrs["label"] = label or ""
                attrs.update({
                    "shape": "box",
                    "fillcolor": parameters.get("transition_fill_color", "white"),
                    "style": "filled",
                    "color": parameters.get("transition_color", "black"),
                    "width": f"{width_gv:.2f}",
                    "height": f"{height_gv:.2f}",
                    "fixedsize": "true",
                    "fontname": parameters.get("font_name", "Arial"),
                    "fontsize": str(font_size),
                    "fontcolor": parameters.get("transition_textcolor", "black"),
                })

        elif node['type'] == DUMMY_TYPE:
            attrs.update({
                "shape": "point",
                "width": "0.01",
                "height": "0.01",
                "label": "",
                "fixedsize": "true",
                "style": "invis",
            })

        viz.node(node_id, **attrs)

    # Add Edges
    arcs = layout_data["arcs"]
    original_arcs = layout_data["original_arcs"]

    direct_edges = []
    waypoint_edges = {}

    for arc_id, arc in arcs.items():
        src = arc['source']
        tgt = arc['target']

        if src not in vertices or tgt not in vertices or \
                vertices[src].get('x') is None or vertices[tgt].get('x') is None:
            continue

        src_type = vertices[src]['type']
        tgt_type = vertices[tgt]['type']

        original_arc_id = arc.get('original_arc_id', arc_id)
        original_arc = original_arcs.get(original_arc_id, arc)

        if src_type != DUMMY_TYPE and tgt_type != DUMMY_TYPE and not original_arc.get('dummy_nodes'):
            direct_edges.append(arc)
        else:
            if original_arc_id not in waypoint_edges:
                ultimate_src = original_arc.get('source')
                ultimate_tgt = original_arc.get('target')
                dummy_nodes = original_arc.get('dummy_nodes', [])
                waypoint_edges[original_arc_id] = {
                    'original_arc': original_arc,
                    'segments': [],
                    'ultimate_source': ultimate_src,
                    'ultimate_target': ultimate_tgt,
                    'dummy_nodes': dummy_nodes
                }
            waypoint_edges[original_arc_id]['segments'].append(arc)

    # Draw direct edges: ALWAYS in original semantic direction (forward)
    for arc in direct_edges:
        src = arc['source']
        tgt = arc['target']
        ot = arc.get("objectType", layout_data["object_types"][0] if layout_data["object_types"] else "unknown")
        ot_color = ot_to_color(ot)

        edge_attrs = {
            "color": ot_color,
            "penwidth": str(parameters.get("arc_size", 1.0) * arc.get("weight", 1.0)),
            "arrowsize": str(parameters.get("arrowhead_size", 0.7)),
            "id": arc['id'],
            "splines": "line",
            "dir": "forward",   # <<--- force forward to avoid inverted arrows
        }

        if arc.get('variable'):
            edge_attrs["penwidth"] = str(float(edge_attrs["penwidth"]) * 1.5)
            edge_attrs["style"] = "dashed"

        viz.edge(src, tgt, **edge_attrs)

    # Draw segmented (dummy) edges with arrow at the original target side
    for orig_arc_id, wp_data in waypoint_edges.items():
        orig_arc = wp_data['original_arc']
        segments = wp_data['segments']
        ultimate_src = wp_data['ultimate_source']
        ultimate_tgt = wp_data['ultimate_target']

        ot = orig_arc.get("objectType", layout_data["object_types"][0] if layout_data["object_types"] else "unknown")
        ot_color = ot_to_color(ot)

        base_edge_attrs = {
            "color": ot_color,
            "penwidth": str(parameters.get("arc_size", 1.0) * orig_arc.get("weight", 1.0)),
            "arrowsize": str(parameters.get("arrowhead_size", 0.7)),
            "splines": "polyline",
        }

        if orig_arc.get('variable'):
            base_edge_attrs["penwidth"] = str(float(base_edge_attrs["penwidth"]) * 1.5)
            base_edge_attrs["style"] = "dashed"

        for segment in segments:
            src = segment['source']
            tgt = segment['target']
            segment_attrs = base_edge_attrs.copy()
            segment_attrs['id'] = segment['id']

            # Put the arrow at the original target side:
            if tgt == ultimate_tgt:
                segment_attrs['arrowhead'] = 'normal'
                segment_attrs['dir'] = 'forward'
            elif src == ultimate_tgt:
                segment_attrs['arrowhead'] = 'normal'
                segment_attrs['dir'] = 'back'
            else:
                segment_attrs['arrowhead'] = 'none'

            viz.edge(src, tgt, **segment_attrs)

    viz.format = image_format.replace("html", "plain-ext")
    return viz


# --- Main Apply Function ---

class Parameters(Enum):
    """ Parameters for the Sugiyama OCPN visualization """
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"  # TB or LR
    ENGINE = "engine"  # dot, neato, fdp
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"

    # Sugiyama specific parameters (matching TS config where possible)
    SOURCES = "sources"  # List of node IDs to force as sources
    SINKS = "sinks"  # List of node IDs to force as sinks
    VERTEX_SEP = "vertex_sep"  # Minimum horizontal distance between nodes
    LAYER_SEP = "layer_sep"  # Minimum vertical distance between layers
    EDGE_SEP = "edge_sep"  # (Not directly used here, vertex_sep dominates)
    BORDER_PADDING = "border_padding"  # Padding around the graph
    OBJECT_CENTRALITY = "object_centrality"  # Dict {ot: centrality_value}
    OBJECT_ATTRACTION = "object_attraction"  # Weight (0-1) for place attraction in barycenter
    OBJECT_ATTRACTION_RANGE_MIN = "object_attraction_range_min"
    OBJECT_ATTRACTION_RANGE_MAX = "object_attraction_range_max"
    MAX_BARYCENTER_ITERATIONS = "max_barycenter_iterations"

    # Visual element sizes (in points, roughly 1/72 inch)
    PLACE_RADIUS = "place_radius"
    TRANSITION_WIDTH = "transition_width"
    TRANSITION_HEIGHT = "transition_height"
    SILENT_TRANSITION_WIDTH = "silent_transition_width"
    ARC_SIZE = "arc_size"
    ARROWHEAD_SIZE = "arrowhead_size"
    INDICATE_ARC_WEIGHT = "indicate_arc_weight"
    INDICATE_VARIABLE_ARCS = "indicate_variable_arcs"
    VARIABLE_ARC_INDICATOR_COLOR = "variable_arc_indicator_color"
    VARIABLE_ARC_INDICATOR_SIZE = "variable_arc_indicator_size"

    # Colors
    TRANSITION_COLOR = "transition_color"
    TRANSITION_FILL_COLOR = "transition_fill_color"
    TRANSITION_TEXT_COLOR = "transition_text_color"
    DEFAULT_PLACE_COLOR = "default_place_color"
    ARC_DEFAULT_COLOR = "arc_default_color"
    TYPE_COLOR_MAPPING = "type_color_mapping"

    # Font
    FONT_NAME = "font_name"
    FONT_SIZE = "font_size"


def apply(ocpn: Dict[str, Any], parameters: Optional[Dict[Any, Any]] = None) -> Digraph:
    """
    Obtains a visualization of the provided object-centric Petri net using
    a Sugiyama-based layout algorithm.

    Args:
        ocpn (Dict[str, Any]): Object-centric Petri net structure.
        parameters (Optional[Dict[Any, Any]], optional): Algorithm parameters.

    Returns:
        Digraph: A Graphviz digraph object representing the layout.
    """
    if parameters is None:
        parameters = {}

    ocpn = deepcopy(ocpn)

    for ot, (net, im, fm) in ocpn.get("petri_nets", {}).items():
        for place in net.places:
            place.name = ot + "@@" + place.name

    # --- Set default configuration parameters ---
    config = {
        # Basic viz
        Parameters.FORMAT.value: constants.DEFAULT_FORMAT_GVIZ_VIEW,
        Parameters.BGCOLOR.value: constants.DEFAULT_BGCOLOR,
        Parameters.RANKDIR.value: constants.DEFAULT_RANKDIR_GVIZ,
        Parameters.ENGINE.value: "dot",
        Parameters.ENABLE_GRAPH_TITLE.value: constants.DEFAULT_ENABLE_GRAPH_TITLES,
        Parameters.GRAPH_TITLE.value: "Object-Centric Petri Net (Sugiyama Layout)",
        # Sugiyama layout params
        Parameters.SOURCES.value: [],
        Parameters.SINKS.value: [],
        Parameters.VERTEX_SEP.value: 30,
        Parameters.LAYER_SEP.value: 50,
        Parameters.BORDER_PADDING.value: 20,
        Parameters.OBJECT_CENTRALITY.value: {},
        Parameters.OBJECT_ATTRACTION.value: 0.2,
        Parameters.OBJECT_ATTRACTION_RANGE_MIN.value: 1,
        Parameters.OBJECT_ATTRACTION_RANGE_MAX.value: 2,
        Parameters.MAX_BARYCENTER_ITERATIONS.value: 24,
        # Visual sizes (points)
        Parameters.PLACE_RADIUS.value: 10.0,
        Parameters.TRANSITION_WIDTH.value: 40.0,
        Parameters.TRANSITION_HEIGHT.value: 20.0,
        Parameters.SILENT_TRANSITION_WIDTH.value: 8.0,
        Parameters.ARC_SIZE.value: 1.0,
        Parameters.ARROWHEAD_SIZE.value: 0.7,
        Parameters.INDICATE_ARC_WEIGHT.value: False,
        Parameters.INDICATE_VARIABLE_ARCS.value: True,
        Parameters.VARIABLE_ARC_INDICATOR_COLOR.value: "#FF0000",
        Parameters.VARIABLE_ARC_INDICATOR_SIZE.value: 1.5,
        # Colors
        Parameters.TRANSITION_COLOR.value: "black",
        Parameters.TRANSITION_FILL_COLOR.value: "white",
        Parameters.TRANSITION_TEXT_COLOR.value: "black",
        Parameters.DEFAULT_PLACE_COLOR.value: "#D3D3D3",
        Parameters.ARC_DEFAULT_COLOR.value: "black",
        Parameters.TYPE_COLOR_MAPPING.value: {},
        # Font
        Parameters.FONT_NAME.value: "Arial",
        Parameters.FONT_SIZE.value: 10,
    }
    config.update({param.value: val for param, val in parameters.items() if isinstance(param, Parameters)})
    config.update({k: v for k, v in parameters.items() if isinstance(k, str) and k in [p.value for p in Parameters]})

    layout_data = _build_initial_layout(ocpn, config)
    if not layout_data["vertices"]:
        print("Error: No vertices found in OCPN structure.")
        return Digraph()

    _reverse_cycles(layout_data, config)

    if not _assign_layers(layout_data):
        print("Error: Layer assignment failed. Cannot proceed.")
        return Digraph()

    _insert_dummy_vertices(layout_data)

    _order_vertices(layout_data, config)

    _position_vertices(layout_data, config)

    gviz_graph = _create_graphviz_digraph(ocpn, layout_data, config)

    return gviz_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import heapq
import sys
import uuid
from enum import Enum

from pm4py.util import constants, xes_constants, exec_utils, pandas_utils, thread_utils
from pm4py.util import variants_util
from pm4py.objects.petri_net.utils import align_utils
from pm4py.objects.log.obj import EventLog, Trace, Event
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.util import typing
from pm4py.objects.conversion.log import converter as log_converter
import pandas as pd


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    SYNC_COST_FUNCTION = "sync_cost_function"
    MODEL_MOVE_COST_FUNCTION = "model_move_cost_function"
    LOG_MOVE_COST_FUNCTION = "log_move_cost_function"
    INTERNAL_LOG_MOVE_COST_FUNCTION = "internal_log_move_cost_function"
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"


class Outputs(Enum):
    ALIGNMENT = "alignment"
    COST = "cost"
    VISITED = "visited_states"
    CLOSED = "closed"
    INTERNAL_COST = "internal_cost"


POSITION_G = 0
POSITION_LOG = 1
POSITION_NUM_VISITED = 2
POSITION_H = 3
POSITION_REAL_F = 4
POSITION_F = 5
POSITION_MODEL = 6
POSITION_IS_LM = 7
POSITION_IS_MM = 8
POSITION_PREV = 9


def apply(
    obj: Union[EventLog, Trace],
    dfg: Dict[Tuple[str, str], int],
    sa: Dict[str, int],
    ea: Dict[str, int],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Union[typing.AlignmentResult, typing.ListAlignments]:
    """
    Applies the alignment algorithm provided a log/trace object, and a *connected* DFG

    Parameters
    --------------
    obj
        Event log / Trace
    dfg
        *Connected* directly-Follows Graph
    sa
        Start activities
    ea
        End activities
    parameters
        Parameters of the algorithm:
        - Parameters.SYNC_COST_FUNCTION: for each activity that is in both the trace and the model, provide the
        non-negative cost of a sync move
        - Parameters.MODEL_MOVE_COST_FUNCTION: for each activity that is in the model, provide the non-negative
        cost of a model move
        - Parameters.LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost of a log move
        that is returned in the alignment to the user (but not used internally for ordering the states)
        - Parameters.INTERNAL_LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost
        of a log move that is used internally for ordering the states in the search algorithm.
        - Parameters.ACTIVITY_KEY: the attribute of the log that is the activity

    Returns
    --------------
    ali
        Result of the alignment
    """
    if isinstance(obj, Trace):
        return apply_trace(obj, dfg, sa, ea, parameters=parameters)
    else:
        return apply_log(obj, dfg, sa, ea, parameters=parameters)


def apply_log(log, dfg, sa, ea, parameters=None):
    """
    Applies the alignment algorithm provided a log object, and a *connected* DFG

    Parameters
    ----------------
    log
        Event log
    dfg
        *Connected* DFG
    sa
        Start activities
    ea
        End activities
    parameters
        Parameters of the algorithm:
        - Parameters.SYNC_COST_FUNCTION: for each activity that is in both the trace and the model, provide the
        non-negative cost of a sync move
        - Parameters.MODEL_MOVE_COST_FUNCTION: for each activity that is in the model, provide the non-negative
        cost of a model move
        - Parameters.LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost of a log move
        that is returned in the alignment to the user (but not used internally for ordering the states)
        - Parameters.INTERNAL_LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost
        of a log move that is used internally for ordering the states in the search algorithm.
        - Parameters.ACTIVITY_KEY: the attribute of the log that is the activity

    Returns
    ----------------
    aligned_traces
        For each trace, contains a dictionary describing the alignment
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    aligned_traces = []
    align_dict = {}

    al_empty_cost = __apply_list_activities(
        [], dfg, sa, ea, parameters=parameters
    )["cost"]

    if pandas_utils.check_is_pandas_dataframe(log):
        case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
        )
        traces = pandas_utils.get_traces(log, case_id_key, activity_key)
    else:
        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_EVENT_LOG,
            parameters=parameters,
        )
        traces = [tuple(x[activity_key] for x in trace) for trace in log]

    unique_traces = set(traces)

    thm = thread_utils.Pm4pyThreadManager()
    f = lambda x, y: __compute_alignments(x, dfg, sa, ea, y, al_empty_cost, parameters)

    for trace_act in unique_traces:
        thm.submit(f, trace_act, align_dict)

    thm.join()

    for trace_act in traces:
        aligned_traces.append(align_dict[trace_act])

    return aligned_traces


def __compute_alignments(trace_act, dfg, sa, ea, align_dict, al_empty_cost, parameters):
    log_move_cost_function = exec_utils.get_param_value(
        Parameters.LOG_MOVE_COST_FUNCTION,
        parameters,
        {x: align_utils.STD_MODEL_LOG_MOVE_COST for x in trace_act},
    )
    trace_bwc_cost = sum(log_move_cost_function[x] for x in trace_act)
    al_tr = __apply_list_activities(
        trace_act, dfg, sa, ea, parameters=parameters
    )
    al_tr["fitness"] = 1.0 - al_tr["cost"] / (
            al_empty_cost + trace_bwc_cost
    )
    al_tr["bwc"] = al_empty_cost + trace_bwc_cost
    align_dict[trace_act] = al_tr


def apply_trace(trace, dfg, sa, ea, parameters=None):
    """
    Applies the alignment algorithm provided a trace of a log, and a *connected* DFG

    Parameters
    ---------------
    trace
        Trace
    dfg
        *Connected* DFG
    sa
        Start activities
    ea
        End activities
    parameters
        Parameters of the algorithm:
        - Parameters.SYNC_COST_FUNCTION: for each activity that is in both the trace and the model, provide the
        non-negative cost of a sync move
        - Parameters.MODEL_MOVE_COST_FUNCTION: for each activity that is in the model, provide the non-negative
        cost of a model move
        - Parameters.LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost of a log move
        that is returned in the alignment to the user (but not used internally for ordering the states)
        - Parameters.INTERNAL_LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost
        of a log move that is used internally for ordering the states in the search algorithm.
        - Parameters.ACTIVITY_KEY: the attribute of the log that is the activity

    Returns
    ---------------
    ali
        Dictionary describing the alignment
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    trace_act = tuple(x[activity_key] for x in trace)

    return __apply_list_activities(
        trace_act, dfg, sa, ea, parameters=parameters
    )


def apply_from_variants_list_dfg_string(
    var_list, dfg_serialization, parameters=None
):
    if parameters is None:
        parameters = {}

    from pm4py.objects.dfg.importer import importer as dfg_importer

    dfg, sa, ea = dfg_importer.deserialize(
        dfg_serialization, parameters=parameters
    )

    return apply_from_variants_list(
        var_list, dfg, sa, ea, parameters=parameters
    )


def apply_from_variants_list(var_list, dfg, sa, ea, parameters=None):
    if parameters is None:
        parameters = {}

    dictio_alignments = {}
    for varitem in var_list:
        variant = varitem[0]
        dictio_alignments[variant] = apply_from_variant(
            variant, dfg, sa, ea, parameters=parameters
        )
    return dictio_alignments


def apply_from_variant(variant, dfg, sa, ea, parameters=None):
    if parameters is None:
        parameters = {}

    trace = variants_util.variant_to_trace(variant, parameters=parameters)
    return apply_trace(trace, dfg, sa, ea, parameters=parameters)


def dijkstra_to_end_node(
    dfg,
    sa,
    ea,
    start_node,
    end_node,
    activities_model,
    sync_cost_function,
    model_move_cost_function,
):
    """
    Gets the cost of the minimum path from a node to the end node

    Parameters
    ---------------
    dfg
        *Connected* DFG
    sa
        Start activities
    ea
        End activities
    start_node
        Start node of the graph (connected to all the start activities)
    end_node
        End node of the graph (connected to all the end activities)
    activities_model
        Set of the activities contained in the DFG
    sync_cost_function
        Given an activity, provides the cost when the activity is executed in a sync way
    model_move_cost_function
        Given an activity, provides the cost when the activity is executed as a move-on-model

    Returns
    -----------------
    cost_to_end_node
        Dictionary associating to each node the cost to the end node
    """
    pred = {x: set() for x in activities_model}
    pred[start_node] = set()
    pred[end_node] = set()
    for x in sa:
        pred[x].add(start_node)
    for x in ea:
        pred[end_node].add(x)
    for x in dfg:
        pred[x[1]].add(x[0])
    min_dist = {x: sys.maxsize for x in activities_model}
    min_dist[start_node] = sys.maxsize
    min_dist[end_node] = 0

    nodes_cost = {}
    nodes_cost[start_node] = 0
    nodes_cost[end_node] = 0

    for x in activities_model:
        if x in sync_cost_function:
            nodes_cost[x] = min(
                sync_cost_function[x], model_move_cost_function[x]
            )
        else:
            nodes_cost[x] = model_move_cost_function[x]

    to_visit = [end_node]
    while len(to_visit) > 0:
        curr = to_visit.pop(0)
        curr_dist = min_dist[curr]
        for prev_node in pred[curr]:
            prev_dist = min_dist[prev_node]
            if prev_dist > curr_dist + nodes_cost[prev_node]:
                min_dist[prev_node] = curr_dist + nodes_cost[prev_node]
                to_visit.append(prev_node)

    # subtract the cost of the node itself to make the heuristics correct
    for el in min_dist:
        min_dist[el] -= nodes_cost[el]

    return min_dist


def __apply_list_activities(trace, dfg, sa, ea, parameters=None):
    """
    Apply the alignments provided the *connected* DFG and a list of activities of the trace

    Parameters
    --------------
    trace
        List of activities of the trace
    dfg
        *Connected* DFG
    sa
        Start activities
    ea
        End activities
    parameters
        Parameters of the algorithm:
        - Parameters.SYNC_COST_FUNCTION: for each activity that is in both the trace and the model, provide the
        non-negative cost of a sync move
        - Parameters.MODEL_MOVE_COST_FUNCTION: for each activity that is in the model, provide the non-negative
        cost of a model move
        - Parameters.LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost of a log move
        that is returned in the alignment to the user (but not used internally for ordering the states)
        - Parameters.INTERNAL_LOG_MOVE_COST_FUNCTION: for each activity that is in the trace, provide the cost
        of a log move that is used internally for ordering the states in the search algorithm.

    Returns
    --------------
    ali
        Dictionary describing the alignment
    """
    if parameters is None:
        parameters = {}

    activities_trace = set(x for x in trace)
    activities_model = (
        set(x[0] for x in dfg)
        .union(set(x[1] for x in dfg))
        .union(set(x for x in sa))
        .union(set(x for x in ea))
    )
    activities_one_instantiation = activities_trace.union(activities_model)
    activities_one_instantiation = {x: x for x in activities_one_instantiation}
    activities_both_in_trace_and_model = activities_trace.intersection(
        activities_model
    )

    sync_cost_function = exec_utils.get_param_value(
        Parameters.SYNC_COST_FUNCTION,
        parameters,
        {
            x: align_utils.STD_SYNC_COST
            for x in activities_both_in_trace_and_model
        },
    )
    model_move_cost_function = exec_utils.get_param_value(
        Parameters.MODEL_MOVE_COST_FUNCTION,
        parameters,
        {x: align_utils.STD_MODEL_LOG_MOVE_COST for x in activities_model},
    )
    # for each activity that is in the trace, provides the cost of a log move
    # that is returned in the alignment to the user (but not used internally
    # for ordering the states)
    log_move_cost_function = exec_utils.get_param_value(
        Parameters.LOG_MOVE_COST_FUNCTION,
        parameters,
        {x: align_utils.STD_MODEL_LOG_MOVE_COST for x in activities_trace},
    )
    # for each activity that is in the trace, provide the cost of a log move that is used internally for
    # ordering the states in the search algorithm.
    internal_log_move_cost_function = exec_utils.get_param_value(
        Parameters.INTERNAL_LOG_MOVE_COST_FUNCTION, parameters, None
    )
    if internal_log_move_cost_function is None:
        internal_log_move_cost_function = {}
        for x in activities_trace:
            if x in activities_model:
                internal_log_move_cost_function[x] = log_move_cost_function[x]
            else:
                internal_log_move_cost_function[x] = 0

    start_node = str(uuid.uuid4())
    end_node = str(uuid.uuid4())

    outgoing_nodes = {
        activities_one_instantiation[x]: set() for x in activities_model
    }
    outgoing_nodes[start_node] = set()
    for x in dfg:
        outgoing_nodes[x[0]].add(activities_one_instantiation[x[1]])
    for x in sa:
        outgoing_nodes[start_node].add(activities_one_instantiation[x])
    for x in ea:
        outgoing_nodes[activities_one_instantiation[x]].add(end_node)

    # closed set: the set associates each activity of the model with the index of the trace that lastly visited that.
    # this is a very efficient closed set since it cuts double-visiting the same activity for state visited later
    # that have a lower index of a trace.
    closed = {activities_one_instantiation[x]: -1 for x in activities_model}
    closed[start_node] = -1
    closed[end_node] = -1

    f = 0
    h = 0
    g = f + h

    num_visited = 0
    num_closed = 0

    open_set = [(g, 0, num_visited, -h, f, f, start_node, False, False, None)]
    heapq.heapify(open_set)

    while not len(open_set) == 0:
        curr = heapq.heappop(open_set)

        num_visited = num_visited + 1

        if -curr[POSITION_LOG] <= closed[curr[POSITION_MODEL]]:
            # closed
            continue

        closed[curr[POSITION_MODEL]] = -curr[POSITION_LOG]
        not_end_trace = -curr[POSITION_LOG] < len(trace)

        if curr[POSITION_MODEL] == end_node and not not_end_trace:
            return __return_alignment(curr, trace, num_closed)

        num_closed = num_closed + 1

        trace_curr = trace[-curr[POSITION_LOG]] if not_end_trace else None

        if not curr[POSITION_MODEL] == end_node:
            # I can do some sync / move on model
            list_act = outgoing_nodes[curr[POSITION_MODEL]]

            act_not_in_model = (
                not_end_trace and trace_curr not in activities_model
            )

            # avoid scheduling model moves if the activity of the log is not in the model,
            # in that case schedule directly the model move
            if not act_not_in_model:
                can_sync = not_end_trace and trace_curr in list_act

                if can_sync:
                    f = curr[POSITION_F] + sync_cost_function[trace_curr]
                    real_f = (
                        curr[POSITION_REAL_F] + sync_cost_function[trace_curr]
                    )
                    h = 0

                    g = f + h

                    new_state = (
                        g,
                        curr[POSITION_LOG] - 1,
                        num_visited,
                        -h,
                        real_f,
                        f,
                        activities_one_instantiation[trace_curr],
                        False,
                        False,
                        curr,
                    )
                    if (
                        -new_state[POSITION_LOG]
                        > closed[new_state[POSITION_MODEL]]
                    ):
                        heapq.heappush(open_set, new_state)

                for act in list_act:
                    if act == end_node:
                        new_state = (
                            curr[POSITION_G],
                            curr[POSITION_LOG],
                            num_visited,
                            curr[POSITION_H],
                            curr[POSITION_REAL_F],
                            curr[POSITION_F],
                            end_node,
                            False,
                            False,
                            curr,
                        )
                        if (
                            -new_state[POSITION_LOG]
                            > closed[new_state[POSITION_MODEL]]
                        ):
                            heapq.heappush(open_set, new_state)
                    elif not not_end_trace or trace_curr not in list_act:
                        f = curr[POSITION_F] + model_move_cost_function[act]
                        real_f = (
                            curr[POSITION_REAL_F]
                            + model_move_cost_function[act]
                        )
                        h = 0

                        g = f + h

                        new_state = (
                            g,
                            curr[POSITION_LOG],
                            num_visited,
                            -h,
                            real_f,
                            f,
                            activities_one_instantiation[act],
                            False,
                            True,
                            curr,
                        )
                        if (
                            -new_state[POSITION_LOG]
                            > closed[new_state[POSITION_MODEL]]
                        ):
                            heapq.heappush(open_set, new_state)

        if not_end_trace and not curr[POSITION_IS_MM]:
            # I can do some move-on-log
            f = curr[POSITION_F] + internal_log_move_cost_function[trace_curr]
            real_f = curr[POSITION_REAL_F] + log_move_cost_function[trace_curr]
            h = 0

            g = f + h

            new_state = (
                g,
                curr[POSITION_LOG] - 1,
                num_visited,
                -h,
                real_f,
                f,
                curr[POSITION_MODEL],
                True,
                False,
                curr,
            )
            if -new_state[POSITION_LOG] > closed[new_state[POSITION_MODEL]]:
                heapq.heappush(open_set, new_state)


def __return_alignment(state, trace, closed):
    """
    Returns the computed alignment in an intellegible way

    Parameters
    --------------
    state
        Current state
    trace
        List of activities
    closed
        Number of closed states

    Returns
    --------------
    ali
        Dictionary describing the alignment
    """
    cost = state[POSITION_REAL_F]
    internal_cost = state[POSITION_F]
    visited = state[POSITION_NUM_VISITED]
    moves = []
    while state[POSITION_PREV] is not None:
        state = state[POSITION_PREV]
        is_lm = state[POSITION_IS_LM]
        is_mm = state[POSITION_IS_MM]
        pl = -state[POSITION_LOG] - 1
        if pl >= 0:
            lm = ">>"
            mm = ">>"
            if not is_lm:
                mm = state[POSITION_MODEL]
            if not is_mm:
                lm = trace[pl]
            moves.append((lm, mm))
    moves.reverse()

    return {
        Outputs.ALIGNMENT.value: moves,
        Outputs.COST.value: cost,
        Outputs.VISITED.value: visited,
        Outputs.CLOSED.value: closed,
        Outputs.INTERNAL_COST.value: internal_cost,
    }


def project_log_on_dfg(
    log: Union[EventLog, pd.DataFrame],
    dfg: Dict[Tuple[str, str], int],
    sa: Dict[str, int],
    ea: Dict[str, int],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Projects the traces of an event log to the specified DFG, in order to assess the conformance of the different
    directly-follows relationships and their performance (as the timestamps are recorded).
    The result is a event log where each 'projected' trace can be replayed on the given DFG.
    Each event of a 'projected' trace has the '@@is_conforming' attribute set to:
    - True when the activity is mimicked by the original trace (sync move)
    - False when the activity is not reflected in the original trace (move-on-model)
    Move-on-log (activities of the trace that are not mimicked by the DFG) are skipped altogether.

    Minimum Viable Example:

        import pm4py
        from pm4py.algo.conformance.alignments.dfg.variants import classic as dfg_alignments

        log = pm4py.read_xes("tests/input_data/receipt.xes", return_legacy_log_object=True)
        filtered_log = pm4py.filter_variants_top_k(log, 1)
        dfg, sa, ea = pm4py.discover_dfg(filtered_log)

        projected_log = dfg_alignments.project_log_on_dfg(log, dfg, sa, ea)
        pm4py.write_xes(projected_log, "projected_log.xes")


    Parameters
    ---------------
    log
        Event log
    dfg
        Directly-Follows Graph
    sa
        Start activities
    ea
        End activities
    parameters
        Parameters of the method, including:
            - Parameters.ACTIVITY_KEY => the attribute of the event log to be used as activity
            - Parameters.TIMESTAMP_KEY => the attribute of the event log to be used as timestamp

    Returns
    ---------------
    projected_log
        Projected event log with the aforementioned features
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    log = log_converter.apply(log, parameters=parameters)

    aligned_traces = apply_log(log, dfg, sa, ea, parameters=parameters)

    projected_log = EventLog()
    projected_log.attributes["@@aligned_dfg"] = repr([dfg, sa, ea])

    for i in range(len(log)):
        projected_trace = Trace()
        projected_trace.attributes["@@original_trace"] = ",".join(
            x[activity_key] for x in log[i]
        )

        if len(log[i]) > 0:
            trace = log[i]
            alignment = aligned_traces[i]["alignment"]
            z = 0
            curr_timestamp = trace[z][timestamp_key]

            for j in range(len(alignment)):
                if alignment[j][1] == ">>":
                    # move on log
                    pass
                elif (
                    alignment[j][0] == ">>"
                    or alignment[j][0] == alignment[j][1]
                ):
                    is_conforming = False
                    if alignment[j][0] != ">>":
                        # sync move
                        curr_timestamp = trace[z][timestamp_key]
                        is_conforming = True
                        z = z + 1
                    event = Event(
                        {
                            activity_key: alignment[j][1],
                            timestamp_key: curr_timestamp,
                            "@@is_conforming": is_conforming,
                        }
                    )
                    projected_trace.append(event)

        projected_log.append(projected_trace)

    return projected_log


def project_alignments_on_dfg(log: Union[EventLog, pd.DataFrame],
                              dfg: Dict[Tuple[str, str], int],
                              sa: Dict[str, int],
                              ea: Dict[str, int],
                              parameters: Optional[Dict[Union[str,
                                                              Parameters], Any]] = None,
                              ) -> Tuple[Dict[Tuple[str, str], Dict[str, int]], Dict[str, Dict[str, int]]]:
    """
    Projects alignment results onto the directly-follows graph (DFG) and activity dictionary.

    This method performs a DFG-based alignment between an event log (or DataFrame) and a process model
    represented by a directly-follows graph. The alignment moves are categorized into three types:

      - SYNC (Synchronous Move): Both the log event and the model activity match.
      - MM (Model Move): The model activity is executed without a corresponding log event.
      - LM (Log Move): A log event occurs without a corresponding model activity.

    After computing the alignments using the provided log and DFG (via the `apply_log` function),
    the method "projects" these results by aggregating:

      - For each DFG edge (transition), the number of SYNC and MM moves.
      - For each activity, the number of SYNC, MM, and LM moves.

    Parameters
    ----------
    log : Union[EventLog, pd.DataFrame]
        The event log or DataFrame containing the events.
    dfg : Dict[Tuple[str, str], int]
        A dictionary representing the directly-follows graph where each key is a tuple of activities
        (from, to) and each value is the frequency of that transition.
    sa : Dict[str, int]
        A dictionary of start activities with their corresponding frequencies.
    ea : Dict[str, int]
        A dictionary of end activities with their corresponding frequencies.
    parameters : Optional[Dict[Union[str, Parameters], Any]], optional
        Additional parameters to control the alignment process, by default None.

    Returns
    -------
    Tuple[Dict[Tuple[str, str], Dict[str, int]], Dict[str, Dict[str, int]]]
        A tuple containing:
          - conformance_dfg: A dictionary where each key is an edge (tuple of activities) from the DFG,
            and the value is a dictionary with keys:
                'sync' : count of synchronous moves on that edge.
                'mm'   : count of model moves on that edge.
          - conformance_activities: A dictionary where each key is an activity and the value is a dictionary
            with keys:
                'sync' : count of synchronous moves for the activity.
                'mm'   : count of model moves for the activity.
                'lm'   : count of log moves for the activity.
    """
    if parameters is None:
        parameters = {}

    aligned_traces = apply_log(log, dfg, sa, ea, parameters=parameters)

    activities = set(x[0] for x in dfg).union(set(x[1] for x in dfg))
    conformance_dfg = {x: {"sync": 0, "mm": 0} for x in dfg}
    activities_conformance = {
        a: {"sync": 0, "mm": 0, "lm": 0} for a in activities}

    for trace in aligned_traces:
        trace = trace["alignment"]

        for i in range(len(trace) - 1):
            move0 = trace[i]
            move1 = trace[i + 1]

            if move0[1] == ">>":
                if move0[0] in activities:
                    activities_conformance[move0[0]]["lm"] += 1
            elif move0[0] == ">>":
                if move0[1] in activities:
                    activities_conformance[move0[1]]["mm"] += 1
            else:
                if move0[0] in activities:
                    activities_conformance[move0[0]]["sync"] += 1

            if i == len(trace) - 1:
                if move1[1] == ">>":
                    if move1[0] in activities:
                        activities_conformance[move1[0]]["lm"] += 1
                elif move1[0] == ">>":
                    if move1[1] in activities:
                        activities_conformance[move1[1]]["mm"] += 1
                else:
                    if move1[0] in activities:
                        activities_conformance[move1[0]]["sync"] += 1

            if (move0[1], move1[1]) in conformance_dfg:
                is_sync = move0[0] != ">>" and move1[0] != ">>"

                if is_sync:
                    conformance_dfg[(move0[1], move1[1])]["sync"] += 1
                else:
                    conformance_dfg[(move0[1], move1[1])]["mm"] += 1

    return conformance_dfg, activities_conformance



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np

from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to


def project_net_on_place(place):
    """
    Project a Petri net on a place

    Parameters
    -------------
    place
        Place

    Returns
    -------------
    net
        (Place) net
    im
        Empty initial marking
    fm
        Empty final marking
    """
    place_net = PetriNet()
    place_net_im = Marking()
    place_net_fm = Marking()

    input_trans = [arc.source for arc in place.in_arcs]
    output_trans = [arc.target for arc in place.out_arcs]

    if len(input_trans) == 0 or len(output_trans) == 0:
        raise Exception("place projection not available on source/sink places")

    input_trans_visible = [trans for trans in input_trans if trans.label]
    output_trans_visible = [trans for trans in output_trans if trans.label]

    if not len(input_trans) == len(input_trans_visible) or not len(
        output_trans
    ) == len(output_trans_visible):
        raise Exception(
            "place projection not available on places that have invisible transitions as preset/postset"
        )

    new_place = PetriNet.Place(place.name)
    place_net.places.add(new_place)

    for trans in input_trans:
        new_trans = PetriNet.Transition(trans.name, trans.label)
        place_net.transitions.add(new_trans)
        add_arc_from_to(new_trans, new_place, place_net)

    for trans in output_trans:
        new_trans = PetriNet.Transition(trans.name, trans.label)
        place_net.transitions.add(new_trans)
        add_arc_from_to(new_place, new_trans, place_net)

    return place_net, place_net_im, place_net_fm


def project_net_on_matrix(net, activities, parameters=None):
    """
    Project a Petri net with:
    - only visible transitions
    - where each place preset/postset is disjoint
    - with unique visible transitions
    on a numeric matrix

    Parameters
    --------------
    net
        Petri net
    activities
        List of activities
    parameters
        Possible parameters of the algorithm
    """
    inv_trans_map = {}
    for trans in net.transitions:
        if not trans.label:
            raise Exception(
                "the project_net_on_matrix works only with Petri net that do not contain invisible transitions"
            )
        if trans.label in inv_trans_map:
            raise Exception(
                "the project_net_on_matrix works only with Petri net that contains unique visible transitions"
            )
        inv_trans_map[trans.label] = trans
    places_matrix = []
    for place in net.places:
        place_repr = [0] * len(activities)
        input_trans_labels = set([arc.source.label for arc in place.in_arcs])
        output_trans_labels = set([arc.target.label for arc in place.out_arcs])
        if len(input_trans_labels.intersection(output_trans_labels)) > 0:
            raise Exception(
                "place has a transition that belongs to both preset and postset"
            )
        for arc in place.in_arcs:
            input_trans_label = arc.source.label
            input_arc_weight = arc.weight
            if input_trans_label in activities:
                place_repr[activities.index(input_trans_label)] = (
                    -input_arc_weight
                )
        for arc in place.out_arcs:
            output_trans_label = arc.target.label
            output_arc_weight = arc.weight
            if output_trans_label in activities:
                place_repr[activities.index(output_trans_label)] = (
                    output_arc_weight
                )
        if min(place_repr) < 0 < max(place_repr):
            places_matrix.append(place_repr)
    if len(places_matrix) == 0:
        raise Exception("no places numeric representation could be found")
    places_matrix = np.transpose(np.asmatrix(places_matrix))
    return places_matrix



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import List, TypeVar, Optional, Dict, Any

from pm4py.algo.discovery.inductive.base_case.abc import BaseCase
from pm4py.algo.discovery.inductive.base_case.empty_log import (
    EmptyLogBaseCaseUVCL,
    EmptyLogBaseCaseDFG,
)
from pm4py.algo.discovery.inductive.base_case.single_activity import (
    SingleActivityBaseCaseUVCL,
    SingleActivityBaseCaseDFG,
)
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructure,
    IMDataStructureUVCL,
    IMDataStructureDFG,
)
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree

T = TypeVar("T", bound=IMDataStructure)
S = TypeVar("S", bound=BaseCase)


class BaseCaseFactory:

    @classmethod
    def get_base_cases(
        cls,
        obj: T,
        inst: IMInstance,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[S]:
        if inst is IMInstance.IM or inst is IMInstance.IMf:
            if type(obj) is IMDataStructureUVCL:
                return [EmptyLogBaseCaseUVCL, SingleActivityBaseCaseUVCL]
        if inst is IMInstance.IMd:
            if type(obj) is IMDataStructureDFG:
                return [EmptyLogBaseCaseDFG, SingleActivityBaseCaseDFG]
        return []

    @classmethod
    def apply_base_cases(
        cls,
        obj: T,
        inst: IMInstance,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[ProcessTree]:
        for b in BaseCaseFactory.get_base_cases(obj, inst):
            r = b.apply(obj, parameters)
            if r is not None:
                return r
        return None



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.algo.discovery.dfg.adapters.pandas.df_statistics import (
    get_concurrent_events_dataframe,
)
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, Tuple
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    STRICT = "strict"


def apply(
    dataframe: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Tuple[str, str], int]:
    """
    Gets the number of times for which two activities have been concurrent in the log

    Parameters
    --------------
    dataframe
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.CASE_ID_KEY => case id
        - Parameters.START_TIMESTAMP_KEY => start timestamp
        - Parameters.TIMESTAMP_KEY => complete timestamp
        - Parameters.STRICT => Determine if only entries that are strictly concurrent
            (i.e. the length of the intersection as real interval is > 0) should be obtained. Default: False

    Returns
    --------------
    ret_dict
        Dictionaries associating to a couple of activities (tuple) the number of times for which they have been
        executed in parallel in the log
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    strict = exec_utils.get_param_value(Parameters.STRICT, parameters, False)

    concurrent_dataframe = get_concurrent_events_dataframe(
        dataframe,
        start_timestamp_key=start_timestamp_key,
        timestamp_key=timestamp_key,
        case_id_glue=case_id_glue,
        activity_key=activity_key,
        strict=strict,
    )

    ret_dict0 = (
        concurrent_dataframe.groupby([activity_key, activity_key + "_2"])
        .size()
        .to_dict()
    )
    ret_dict = {}

    # assure to avoid problems with np.float64, by using the Python float type
    for el in ret_dict0:
        # avoid getting two entries for the same set of concurrent activities
        el2 = tuple(sorted(el))
        ret_dict[el2] = int(ret_dict0[el])

    return ret_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Extracts for each object the minimum and the maximum value of the features for the events related to the object.

    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters

    Returns
    -----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    from pm4py.algo.transformation.ocel.features.events import (
        algorithm as event_based_features,
    )

    data_events, feature_names_events = event_based_features.apply(
        ocel, parameters=parameters
    )
    dct_dct_events = event_based_features.transform_features_to_dict_dict(
        ocel, data_events, feature_names_events, parameters=parameters
    )

    ordered_objects = (
        parameters["ordered_objects"]
        if "ordered_objects" in parameters
        else ocel.objects[ocel.object_id_column].to_numpy()
    )

    stream = ocel.relations[
        [ocel.event_id_column, ocel.object_id_column]
    ].to_dict("records")
    obj_rel_evs = {}

    for cou in stream:
        if cou[ocel.object_id_column] not in obj_rel_evs:
            obj_rel_evs[cou[ocel.object_id_column]] = []
        obj_rel_evs[cou[ocel.object_id_column]].append(
            dct_dct_events[cou[ocel.event_id_column]]
        )

    data = []
    feature_names = []
    for x in feature_names_events:
        feature_names.append("@@rel_eve_min_" + x)
        feature_names.append("@@rel_eve_max_" + x)

    for obj in ordered_objects:
        arr = []
        for x in feature_names_events:
            if obj in obj_rel_evs:
                min_v = float(min(y[x] for y in obj_rel_evs[obj]))
                max_v = float(max(y[x] for y in obj_rel_evs[obj]))
            else:
                min_v = 0.0
                max_v = 0.0
            arr.append(min_v)
            arr.append(max_v)
        data.append(arr)

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.util import constants
from typing import Optional, Dict, Any, Union, Tuple


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    KEEP_FIRST_FOLLOWING = "keep_first_following"


def apply(
    interval_log: IMDataStructureUVCL,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Tuple[str, str], int]:
    if parameters is None:
        parameters = {}

    ret_dict = {}
    for trace, freq in interval_log.data_structure.items():
        i = 0
        while i < len(trace):
            act1 = trace[i]
            j = i + 1
            while j < len(trace):
                act2 = trace[j]
                tup = (act1, act2)
                if tup in ret_dict.keys():
                    ret_dict[tup] = ret_dict[tup] + freq
                else:
                    ret_dict[tup] = freq
                j = j + 1
            i = i + 1

    return ret_dict



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import shutil
import tempfile
from enum import Enum
from typing import Any, Dict, Optional

from graphviz import Digraph
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4py.visualization.common.gview import serialize, serialize_dot

from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.util import constants, exec_utils, vis_utils
from pm4py.visualization.heuristics_net.variants import pydotplus_vis


class Variants(Enum):
    PYDOTPLUS = pydotplus_vis


DEFAULT_VARIANT = Variants.PYDOTPLUS


def apply(
    heu_net: HeuristicsNet,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> str:
    """
    Gets a representation of an Heuristics Net

    Parameters
    -------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm, including:
            - Parameters.FORMAT
    variant
        Variant of the algorithm to use:
             - Variants.PYDOTPLUS

    Returns
    ------------
    gviz
        Representation of the Heuristics Net
    """
    return exec_utils.get_variant(variant).apply(
        heu_net, parameters=parameters
    )


def view(figure):
    """
    View on the screen a figure that has been rendered

    Parameters
    ----------
    figure
        figure
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        try:
            filename = figure.name
            figure = filename
        except AttributeError:
            # continue without problems, a proper path has been provided
            pass

        if constants.DEFAULT_GVIZ_VIEW == "matplotlib_view":
            import matplotlib.pyplot as plt
            import matplotlib.image as mpimg

            img = mpimg.imread(figure)
            plt.axis("off")
            plt.tight_layout(pad=0, w_pad=0, h_pad=0)
            plt.imshow(img)
            plt.show()
            return

        if vis_utils.check_visualization_inside_jupyter():
            vis_utils.view_image_in_jupyter(figure)
        else:
            vis_utils.open_opsystem_image_viewer(figure)


def save(gviz: Digraph, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz: Digraph, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: Digraph, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """

    return gview.matplotlib_view(gviz, parameters=parameters)




'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
import polars as pl
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, Union, Tuple


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    STRICT = "strict"


def get_concurrent_events_dataframe(
    lf: pl.LazyFrame,
    start_timestamp_key=None,
    timestamp_key="time:timestamp",
    case_id_glue="case:concept:name",
    activity_key="concept:name",
    strict=False,
) -> pl.LazyFrame:
    """
    Gets the concurrent events (of the same case) in a Polars LazyFrame

    Parameters
    --------------
    lf
        LazyFrame
    start_timestamp_key
        Start timestamp key (if not provided, defaulted to the timestamp_key)
    timestamp_key
        Complete timestamp
    case_id_glue
        Column of the dataframe to use as case ID
    activity_key
        Activity key
    strict
        Gets only entries that are strictly concurrent (i.e. the length of the intersection as real interval is > 0)

    Returns
    ---------------
    conc_ev_dataframe
        Concurrent events LazyFrame with concurrency information
    """
    # if not differently specified, set the start timestamp key to the timestamp key
    if start_timestamp_key is None:
        start_timestamp_key = xes_constants.DEFAULT_START_TIMESTAMP_KEY
        lf = lf.with_columns(pl.col(timestamp_key).alias(start_timestamp_key))

    # Sort by case ID and timestamps
    lf = lf.sort([case_id_glue, start_timestamp_key, timestamp_key])
    
    # Reduce to needed columns for efficiency
    lf = lf.select([case_id_glue, activity_key, start_timestamp_key, timestamp_key])
    
    # Add index for ordering within cases
    lf = lf.with_row_count("__index__")
    
    # Self-join to get all pairs within same cases
    lf_self = lf.select([
        pl.col(case_id_glue).alias(case_id_glue + "_2"),
        pl.col(activity_key).alias(activity_key + "_2"),
        pl.col(start_timestamp_key).alias(start_timestamp_key + "_2"),
        pl.col(timestamp_key).alias(timestamp_key + "_2"),
        pl.col("__index__").alias("__index___2")
    ])
    
    # Join on case ID
    concurrent_df = lf.join(
        lf_self, 
        left_on=case_id_glue, 
        right_on=case_id_glue + "_2",
        how="inner"
    )
    
    # Filter to avoid duplicates - only keep pairs where first index < second index
    concurrent_df = concurrent_df.filter(
        pl.col("__index__") < pl.col("__index___2")
    )
    
    # Calculate overlap of time intervals
    max_start_col = "__max_start_column"
    min_complete_col = "__min_complete_column"
    diff_col = "__diff_maxs_minc"
    
    concurrent_df = concurrent_df.with_columns([
        pl.max_horizontal([pl.col(start_timestamp_key), pl.col(start_timestamp_key + "_2")]).alias(max_start_col),
        pl.min_horizontal([pl.col(timestamp_key), pl.col(timestamp_key + "_2")]).alias(min_complete_col)
    ])
    
    # Calculate time difference (in seconds)
    concurrent_df = concurrent_df.with_columns(
        (pl.col(min_complete_col) - pl.col(max_start_col)).dt.total_seconds().alias(diff_col)
    )
    
    # Filter based on strict parameter
    if strict:
        concurrent_df = concurrent_df.filter(pl.col(diff_col) > 0)
    else:
        concurrent_df = concurrent_df.filter(pl.col(diff_col) >= 0)
    
    return concurrent_df


def apply(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Tuple[str, str], int]:
    """
    Gets the number of times for which two activities have been concurrent in the log

    Parameters
    --------------
    lf
        Polars LazyFrame
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.CASE_ID_KEY => case id
        - Parameters.START_TIMESTAMP_KEY => start timestamp
        - Parameters.TIMESTAMP_KEY => complete timestamp
        - Parameters.STRICT => Determine if only entries that are strictly concurrent
            (i.e. the length of the intersection as real interval is > 0) should be obtained. Default: False

    Returns
    --------------
    ret_dict
        Dictionaries associating to a couple of activities (tuple) the number of times for which they have been
        executed in parallel in the log
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, None
    )
    strict = exec_utils.get_param_value(Parameters.STRICT, parameters, False)

    concurrent_dataframe = get_concurrent_events_dataframe(
        lf,
        start_timestamp_key=start_timestamp_key,
        timestamp_key=timestamp_key,
        case_id_glue=case_id_glue,
        activity_key=activity_key,
        strict=strict,
    )

    # Group by activity pairs and count
    ret_dict0_df = concurrent_dataframe.group_by([activity_key, activity_key + "_2"]).count().collect()
    
    ret_dict0 = {}
    for row in ret_dict0_df.iter_rows():
        key = (row[0], row[1])
        count = row[2]
        ret_dict0[key] = count

    ret_dict = {}

    # Ensure we avoid problems with duplicates by using sorted tuples
    for el in ret_dict0:
        # avoid getting two entries for the same set of concurrent activities
        el2 = tuple(sorted(el))
        if el2 in ret_dict:
            ret_dict[el2] += int(ret_dict0[el])
        else:
            ret_dict[el2] = int(ret_dict0[el])

    return ret_dict


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from enum import Enum
from typing import Optional, Dict, Any, Union, Tuple

from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.util import exec_utils, xes_constants, constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def __is_allowed_prefix(exiting_activities, sa, prefix):
    if not prefix:
        return True
    if prefix[0] not in sa:
        return False
    prev_act = prefix[0]
    for i in range(1, len(prefix)):
        curr_act = prefix[i]
        if (
            prev_act not in exiting_activities
            or curr_act not in exiting_activities[prev_act]
        ):
            return False
        prev_act = curr_act
    if not prefix[-1] in exiting_activities:
        return False
    return True


def apply(
    log: Union[EventLog, EventStream],
    dfg: Dict[Tuple[str, str], int],
    start_activities: Dict[str, int],
    end_activities: Dict[str, int],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Computes the precision of a directly-follows graph using the ETConformance approach

    Parameters
    ---------------
    log
        Event log
    dfg
        Directly-follows graph
    start_activities
        Start activities
    end_activities
        End activities
    parameters
        Parameters of the algorithm:
        - Parameters.ACTIVITY_KEY: the key to use

    Returns
    ----------------
    precision
        Precision value
    """
    if parameters is None:
        parameters = {}
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    precision = 1.0
    sum_ee = 0
    sum_at = 0
    exiting_activities = {}
    for act_couple in dfg:
        if not act_couple[0] in exiting_activities:
            exiting_activities[act_couple[0]] = set()
        exiting_activities[act_couple[0]].add(act_couple[1])
    prefixes = {}
    prefixes_count = Counter()
    for trace in log:
        prefix_act = []
        for i in range(len(trace)):
            prefix_act_tuple = tuple(prefix_act)
            if prefix_act_tuple not in prefixes:
                prefixes[prefix_act_tuple] = set()
            prefixes_count[prefix_act_tuple] += 1
            prefixes[prefix_act_tuple].add(trace[i][activity_key])
            prefix_act.append(trace[i][activity_key])
    for prefix in prefixes:
        if __is_allowed_prefix(exiting_activities, start_activities, prefix):
            log_transitions = prefixes[prefix]
            activated_transitions = (
                set(start_activities.keys())
                if not prefix
                else exiting_activities[prefix[-1]]
            )
            escaping_edges = activated_transitions.difference(log_transitions)
            sum_ee += len(escaping_edges) * prefixes_count[prefix]
            sum_at += len(activated_transitions) * prefixes_count[prefix]

    if sum_at > 0:
        precision = 1 - float(sum_ee) / float(sum_at)

    return precision



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Union, Tuple, Set


class Parameters(Enum):
    ALPHA = "alpha"
    BETA = "beta"
    GAMMA = "gamma"
    THETA_REL = "theta_real"


def apply(real: Union[OCEL,
                      Tuple[Set[str],
                            Set[str],
                            Set[Tuple[str,
                                      str]],
                            Dict[Tuple[str,
                                       str],
                                 int]]],
          normative: Tuple[Set[str],
                           Set[str],
                           Set[Tuple[str,
                                     str]],
                           Dict[Tuple[str,
                                      str],
                                int]],
          parameters: Optional[Dict[Any,
                                    Any]] = None) -> Dict[str,
                                                          Any]:
    """
    Applies ET-OT-based conformance checking between a 'real' object (either an OCEL or an ET-OT graph),
    and a normative ET-OT graph.

    Published in: https://publications.rwth-aachen.de/record/1014107

    Parameters
    -------------------
    real
        Real object (OCEL, or ET-OT graph)
    normative
        Normative object (ET-OT graph)
    parameters
        Variant-specific parameters, including:
        - Parameters.ALPHA
        - Parameters.BETA
        - Parameters.GAMMA
        - Parameters.THETA_REAL

    Returns
    ------------------
    diagn_dict
        Diagnostics dictionary
    """
    if parameters is None:
        parameters = {}

    alpha = exec_utils.get_param_value(Parameters.ALPHA, parameters, 1)
    beta = exec_utils.get_param_value(Parameters.BETA, parameters, 1)
    gamma = exec_utils.get_param_value(Parameters.GAMMA, parameters, 1)
    theta_rel = exec_utils.get_param_value(
        Parameters.THETA_REL, parameters, 0.1)

    if isinstance(real, OCEL):
        from pm4py.algo.discovery.ocel.etot import algorithm as etot_discovery
        real = etot_discovery.apply(real, parameters=parameters)

    return compute_conformance(
        real,
        normative,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        theta_rel=theta_rel)


def compute_conformance(G_L, G_M, alpha=1, beta=1, gamma=1, theta_rel=0.1):
    A_L, OT_L, R_L, w_L = G_L
    A_M, OT_M, R_M, w_M = G_M

    # Node Conformance
    A_missing = A_M - A_L
    A_additional = A_L - A_M
    OT_missing = OT_M - OT_L
    OT_additional = OT_L - OT_M

    # Edge Conformance
    R_missing = R_M - R_L
    R_additional = R_L - R_M

    # Edge Frequency Conformance
    delta_rel_total = 0
    delta_rel = {}
    for r in R_M.intersection(R_L):
        w_M_r = w_M[r]
        w_L_r = w_L[r]
        delta = abs(w_L_r - w_M_r) / w_M_r
        delta_rel[r] = delta
        if delta > theta_rel:
            delta_rel_total += 1

    # Normalization constant
    N = alpha * (len(A_M) + len(OT_M)) + beta * len(R_M) + gamma * len(R_M)

    # Compute numerator
    numerator = alpha * (len(A_missing) + len(OT_missing)) + \
        beta * len(R_missing) + gamma * delta_rel_total

    # Fitness value
    phi = 1 - (numerator / N)

    # Details dictionary
    details = {
        'A_missing': A_missing,
        'A_additional': A_additional,
        'OT_missing': OT_missing,
        'OT_additional': OT_additional,
        'R_missing': R_missing,
        'R_additional': R_additional,
        'delta_rel': delta_rel
    }

    return {"fitness": phi, "details": details}



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.organizational_mining.roles.common import algorithm
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.util import xes_constants as xes
from collections import Counter
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Union, List
from pm4py.objects.log.obj import EventLog
from pm4py.objects.org.roles.obj import Role


class Parameters(Enum):
    ROLES_THRESHOLD_PARAMETER = "roles_threshold_parameter"
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Role]:
    """
    Gets the roles (group of different activities done by similar resources)
    out of the log

    Parameters
    -------------
    log
        Log object
    parameters
        Possible parameters of the algorithm

    Returns
    ------------
    roles
        List of different roles inside the log
    """
    if parameters is None:
        parameters = {}

    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, xes.DEFAULT_RESOURCE_KEY
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )

    stream = log_converter.apply(
        log,
        variant=log_converter.TO_EVENT_STREAM,
        parameters={"deepcopy": False, "include_case_attributes": False},
    )

    activity_resource_couples = Counter(
        (event[resource_key], event[activity_key]) for event in stream
    )

    return algorithm.apply(activity_resource_couples, parameters=parameters)
